from ..exceptions import *

def format_filter_value(type, value):
    """Takes in a filter type and the value to filter for. Depenining on the type 
    the value is formatted with the correct wild card e.g. 'contains' will add a 
    wildcard to the start and end of value *value*

    Args:
        type (str): Either 'equal to', 'contains', 'starts with'
        value (str): value to filter for

    Raises:
        OAUnsupportedFilter: Raised when unsupported filter is requested such as 'ends with' 

    Returns:
        str: Formatted filter with the value and correct filter wildcards.
    """
    if type.lower() == "equal to":
        return f"{value}" 
    elif type.lower() == "starts with":
        return f"{value}*" 
    elif type.lower() == "contains":
        return f"*{value}*"
    else:
        raise OSUnsupportedFilter
    
    
def format_int_list_of_numbers(counrty_code: int, numbers: list):
    """Takes a list of integer numbers with no country code and the country code needed.
    This will then return a list of strings with the country code inserted infront of all
    numbers in orginal list.

    Args:
        counrty_code (int): Country code to be added infront of all numbers in list
        numbers (list): List of integer numbers with no country code

    Returns:
        list: List of numbers in string format with the country code added. 
    """
    numbers.sort()
    formatted_numbers = [f'+{str(counrty_code)}-{str(num)}' for num in numbers]
    return formatted_numbers


import os
import logging 
from logging import Formatter


class Logger:
    
    __instance = None

    # Singleton implementation
    @staticmethod
    def get_instance(user: str):
        if Logger.__instance is None:
            Logger(user)
        return Logger.__instance
    
    def __init__(self, user: str) -> None:
        """Logger class logs API calls made using Odin's Spear.
        Details logged are: Date/ Time, Name (Odin's Spear), User (You the developer), response code, endpoint. 
        Logs are sent to os.log by default but can be changed in set_up_file_handler() method. 
        Logs can also be sent to an external syslog server using the set_up_sys_log_handler method.
        
        NOTE: This object is a singleton and can't be instatiated more than once.

        Args:
            user (str): Username used when creating API object.
        """

        if Logger.__instance is not None:
            raise Exception("Singleton cannot be instantiated more than once!")
        else:
            
            self._user = user
            
            #Remove handlers from root logger
            for handler in logging.root.handlers[:]:
                if isinstance(handler, logging.StreamHandler):
                    logging.root.removeHandler(handler)
            
            self._format = Formatter(fmt="%(asctime)s | %(name)s | %(message)s | %(response_code)s | %(endpoint)s")
            
            self._logger_obj = logging.getLogger("Odin's Spear")
            self._logger_obj.setLevel(logging.DEBUG)
            
            self.set_up_file_handler()
            
            Logger.__instance = self
    
    
    def set_up_file_handler(self, path='os.log'):
        """Creates file handler for Logger object. This defaults to 'os.log' file but this 
        can be changed by passing in a path.

        Args:
            path (str, optional): Path to file to send logs. Defaults to 'os.log'.
        """
        
        from logging import FileHandler
        
        try:
            file_handler = FileHandler(path)
            if not os.path.exists(path):
                os.makedirs(path)
        except Exception:
            print(f'You do not have permission create {path}. Logs saved to os.logg')
            path = 'os.log'
            file_handler = FileHandler(path)
        
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(self._format)
        
        self._logger_obj.addHandler(file_handler)
                
        
    def set_up_sys_log_handler(self, url: str, port_number: int):
        """Sets up a Syslog Server handler which will send logs generated to syslog server. 
        URL to server as well as the port number is needed to send logs. 

        Args:
            url (str): IP/ URL to syslog server e.g. 1.1.1.1 or mysyslogserver.com/
            port_number (int): Port number the logs will sent on to URL
        """
        
        from logging.handlers import SysLogHandler

        self.url = url
        self.port_number = port_number
        
        syslog_handler = SysLogHandler(address=(url, port_number))
        syslog_handler.setLevel(logging.DEBUG)
        syslog_handler.setFormatter(self._format)
        
        self._logger_obj.addHandler(syslog_handler)
        
        
    def _log_request(self, **kwargs):
        """Takes in the response call and response details from Requester 
        and logs this information with logging level INFO.

        Args:
            kwargs: Two keyword arguements: endpoint and response_code
        """        
        
        self._logger_obj.info(
            self._user,
            extra= kwargs
        )
        



def main():
    """ Takes in list of users and performs below:

    Enables VM, 
    Adds VM to email if required along with email address, 
    Transfer on zero (if required) and destination, 
    VM common password, 
    Auto-login

    :param x:
    :param y:
    :param z:

    :return r: 
    """
    return  




class Post():
    
    def __init__(self, requester):
        self.requester = requester
            
#SESSION
        
    def session(self, username, password):
        
        endpoint = "/auth/token"
        
        data={
            "username": username,
            "password": password
        }
        
        return self.requester.post(endpoint, data=data)
        
    def session_switch(self, username: str):
        
        endpoint = "/auth/switch-user"
        
        data = {
            "username": username
        }
        
        return self.requester.post(endpoint, data=data)
        

#ACCOUNT AUTHORIZATION CODES
#ADMINISTRATORS

    def group_admin(self, service_provider_id: str, group_id: str, user_id: str, password: str, payload: dict = {}):
        """Builds a group-level administrator.

        Args:
            service_provider_id (str): Service provider ID where the admin should be built.
            group_id (str): Group ID where the admin should be built
            user_id (str): User ID of the admin. 
            password (str): Password for the administrator profile. Note get.password_generate() can be used to get this.
            payload (dict, optional): Admin configuration data. 
        
        Returns:
            Dict: Returns the admin profile. 
        """

        endpoint = "/groups/admins"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["userId"] = user_id
        payload["password"] = password

        return self.requester.post(endpoint, data=payload)
    
    
    def group_admin_policies_bulk(self, user_ids: list, policy_config: dict):
        """Applies policy settings to multiple group administrators.

        Note: See docs for formatting of parameters.

        Args:
            user_ids (list): User IDs of admins to apply policy to.
            policy_config (dict): Policy settings to apply to target users.

        Returns:
            Dict: Returns admins and policy applied. 
        """
        endpoint = '/groups/admins/policies/bulk'
        
        data = {
            'users': [{'userId':user} for user in user_ids],
            'data': policy_config
        }
        
        self.requester.post(endpoint, data=data)

#ADVICE OF CHARGE
#ALTERNATE NUMBERS
#ANSWER CONFIRMATION
#ALTERNATE USER ID
#ANNOUNCEMENTS
#ANONYMOUS CALL REJECTION
#ATTENDANT CONSOLE
#AUTHENTICATION
#AUTO ATTENDANTS

    def auto_attendant(self, service_provider_id: str, group_id: str, service_user_id: str, aa_name: str, aa_type: str ="Basic", payload: dict ={}):
        """Builds an Auto Attendant (AA) from the given payload.

        Args:
            service_provider_id (str): Service Provider ID of the group where the AA should be built.
            group_id (str): Group ID where the AA should be built.
            service_user_id (str): Service User ID of the AA (including the domain). 
            aa_name (str): Name of the AA
            aa_type (str, optional): Type of AA: "Basic" or "Standard". Will default to "Basic". NOTE: The "Auto Attendant - Standard" service must be enabled on the group in order for the aa_type to be set to "Standard".
            payload (dict, optional): Additional AA configuration data.

        Returns:
            Dict: Returns the AA profile.
        """

        endpoint = "/groups/auto-attendants"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["serviceUserId"] = service_user_id
        payload["serviceInstanceProfile"]["name"] = aa_name
        payload["type"] = aa_type

        default_payload_values = {
            "enableVideo": "false", 
            "extensionDialingScope":"Group",
            "nameDialingScope":"Group",
            "nameDialingEntries":"LastName + FirstName",
            "firstDigitTimeoutSeconds":1
            }

        for key, default_value in default_payload_values.items():
            payload.setdefault(key, default_value)
        
        return self.requester.post(endpoint, data=payload)

    
    def auto_attendant_remove_user(self, service_provider_id: str, group_id: str, user_id: str):
        """Returns a list of the available Auto Attendants (AAs) built in the same group as the specified user.

        Args:
            service_provider_id (str): Service Provider ID where the user is built. 
            group_id (str): Group ID where the user is built. 
            user_id (str): User ID of the user.

        Returns:
            List: List of the Service User IDs of the AAs in the group.
        """

        endpoint = "/groups/auto-attendants/removeUser"

        payload = {
            "serviceProviderId": service_provider_id, 
            "groupId": group_id, 
            "userId": user_id
        }

        return self.requester.post(endpoint, data=payload)
    

    def auto_attendant_submenu(self, service_user_id: str, submenu_id: str, announcement_selection: str="Default", extension_dialing: bool= True):
        """Posts a new submenu to the specified Auto Attendant (AA).

        Args:
            service_user_id (str): Service User ID of the AA.
            submenu_id (str): ID of the submenu to be created. 
            announcement_selection (str, optional): "Default" or "Personal". Defaults to "Default".
            extension_dialing (bool, optional): Whether Level Extension Dialing is enabled or not. Defaults to True.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/auto-attendants/submenus"

        payload = {
            "serviceUserId": service_user_id, 
            "submenuId": submenu_id, 
            "announcementSelection": announcement_selection,
            "enableLevelExtensionDialing": extension_dialing
        }

        return self.requester.post(endpoint, data=payload)


#AUTOMATIC CALLBACK
#AUTOMATIC HOLD RETRIEVE
#BARGE IN EXEMPT
#BASIC CALL LOGS
#BROADWORKS ANYWHERE
#BROADWORKS MOBILITY
#BROADWORKS NAVIGATION
#BROADWORKS RECEPTIONIST ENTERPRISE
#BROADWORKS RECEPTIONIST OFFICE
#BROADWORKS RECEPTIONIST SMALL BUSINESS
#BUSY LAMP FIELD
#CALL CAPACITY
#CALL CENTER
#CALL CONTROL
#CALL FORWARDING ALWAYS
#CALL FORWARDING ALWAYS SECONDARY
#CALL FORWARDING BUSY
#CALL FORWARDING NO ANSWER
#CALL FORWARDING NOT REACHABLE
#CALL FORWARDING SELECTIVE
#CALL FORWARDING SETTINGS
#CALL NOTIFY
#CALL PARK
#CALL PICKUP
#CALL POLICIES
#CALL PROCESSING POLICIES
#CALL RECORDING
#CALL RECORDS
#CALL TRANSFER
#CALL WAITING
#CALLING LINE ID BLOCKING OVERRIDE
#CALLING LINE ID DELIVERY BLOCKING
#CALLING NAME DELIVERY
#CALLING NAME RETRIEVAL
#CALLING NUMBER DELIVERY
#CALLING PARTY CATEGORY
#CALLING PLANS
#CALLBACKS
#CHARGENUMBER
#CLASSMARK
#CLONE
#COLLABORATE
#COMM PILOT CALL MANAGER
#COMM PILOT EXPRESS
#COMMON PHONE LIST
#COMMUNICATION BARRING
#COMMUNICATION BARRING USER
#CONNECTED LINE IDENTIFICATION
#CUSTOM CONTACT DIRECTORY
#DEPARTMENTS
#DEVICE POLICIES
#DEVICES

    def group_device(self, service_provider_id: str, group_id: str, device_name: str, device_type: str, payload: dict ={}):
        """Adds a new device to a group. 

        Args:
            service_provider_id (str): Service provider ID where the device should be built.
            group_id (str): Group ID where the device should be built
            device_name (str): Name of the new device
            device_type (str): Type of device. 
            payload (dict, optional): Device configuration data. 
        
        Returns:
            Dict: Returns the device profile. 
        """
        
        endpoint = "/groups/devices"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["deviceName"] = device_name
        payload["deviceType"] = device_type

        return self.requester.post(endpoint, data=payload)

#DIAL PLAN POLICY
#DIRECTED CALL PICKUP WITH BARGE IN
#DIRECTROUTE
#DN

    def group_dns(self, service_provider_id: str, group_id: str, 
                  start_of_range_number: str, end_of_range_number: str):
        """Adds a range of numbers to a Group. Range of numbers must be complete and 
        format of number must follow: +{country code}-{number}.
        
        Adding a singular number - Set both the start and end of range parameters as the same number.
       
        Args:
            service_provider_id (str): Service provider ID where the target group is located. 
            group_id (str): Group ID where numbers should be added to.
            start_of_range_number (str): Starting number in range to add to group. 
            end_of_range_number (str): Ending number in range to add to group. 
            
        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/dns"
        
        data = {
			"serviceProviderId": service_provider_id,
            "groupId": group_id,
			"dns": [
                {
                    "min": start_of_range_number,
                    "max": end_of_range_number
                }
            ]
		}
        
        return self.requester.post(endpoint, data=data)
    
    
    def group_dns_assign_bulk(self, service_provider_id: str, group_id: str, 
                              start_of_range_number: int, end_of_range_number: int):
        """Adds a range of numbers to a Group. Range of numbers must be complete and 
        format of number must follow: +{country code}-{number}.
        
        Adding a singular number - Set both the start and end of range parameters as the same number.
       
        Args:
            service_provider_id (str): Service provider ID where the target group is located. 
            group_id (str): Group ID where numbers should be added to.
            start_of_range_number (str): Starting number in range to add to group. 
            end_of_range_number (str): Ending number in range to add to group. 
            
        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/dns/assign/bulk"
        
        data = {
			"serviceProviderId": service_provider_id,
            "groupId": group_id,
			"dns": [
                {
                    "min": start_of_range_number,
                    "max": end_of_range_number
                }
            ]
		}
        
        return self.requester.post(endpoint, data=data)
    
    
    def group_dns_unassign_bulk(self, service_provider_id: str, group_id: str, 
                                start_of_range_number: int, end_of_range_number: int):
        """Unassign a range of numbers from a Group. Range of numbers must be complete and 
        format of number must follow: +{country code}-{number}.
        
        Unassigning a singular number - Set both the start and end of range parameters as the same number.
       
        Args:
            service_provider_id (str): Service provider ID where the target group is located. 
            group_id (str): Group ID where numbers where numbers are located.
            start_of_range_number (str): Starting number in range to unassign in group. 
            end_of_range_number (str): Ending number in range to unassign in group.
            
        Returns:
            None: This method does not return any specific value. 
        """
        
        endpoint = "/groups/dns/unassign/bulk"
        
        data = {
			"serviceProviderId": service_provider_id,
            "groupId": group_id,
			"dns": [
                {
                    "min": start_of_range_number,
                    "max": end_of_range_number
                }
            ]
		}
        
        return self.requester.post(endpoint, data=data)
    
    
    def service_provider_dns(self, service_provider_id: str, start_of_range_number: int, 
                             end_of_range_number: int):
        """Adds a range of numbers to a Service Provider/ Enterprise. Range of numbers must be complete and 
        format of number must follow: +{country code}-{number}.
        
        Adding a singular number - Set both the start and end of range parameters as the same number.
       
        Args:
            service_provider_id (str): Service provider ID where the target group is located. 
            start_of_range_number (str): Starting number in range to add to group. 
            end_of_range_number (str): Ending number in range to add to group. 
            
        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/dns"
        
        data = {
			"serviceProviderId": service_provider_id,
			"dns": [
                {
                    "min": start_of_range_number,
                    "max": end_of_range_number
                }
            ]
		}
        
        return self.requester.post(endpoint, data=data)


#DO NOT DISTURB
#DOMAINS
#EMERGENCY NOTIFICATIONS
#EMERGENCY ZONES

    def group_emergency_zones(self, service_provider_id: str, group_id: str, ip_addresses: list):
        """Updates the IP address(es) for the Emergency Zone configured in the group. 
       
        Args:
            service_provider_id (str): Service provider ID where the Emergency Zone to be updated exists.
            group_id (str): Group ID where the Emergency Zone to be updated exists.
            ip_addresses (list): A list of IP address ranges (dicts) to be added to the Emergency Zone. If the IP address to be applied is not a range, the min and max values should be the same.
            
        Returns:
            Dict: Emergency Zone profile with updated IP addresses.
        """

        endpoint = "/groups/emergency-zones"

        data = {
            "serviceProviderId": service_provider_id, 
            "groupId": group_id, 
            "ipAddresses": ip_addresses
        }

        return self.requester.post(endpoint, data=data)


#ENTERPRISE TRUNKS
#EXECUTIVE
#EXECUTIVE ASSISTANT
#EXTENSIONS
#EXTERNAL CALLING LINE ID DELIVERY
#EXTERNAL CUSTOM RINGBACK
#FAX MESSAGING
#FEATURE ACCESS CODES
#FLEXIBLE SEATING
#GROUP PAGING
#GROUPS
#GROUP NAVIGATION
#HOTELING GUEST
#HOTELING HOST
#HUNT GROUPS

    def group_hunt_group(self, service_provider_id: str, group_id: str, service_user_id: str, 
                         clid_first_name: str, clid_last_name:str, extension: str,  payload: dict={}, agents: list=[],
                         policy: str ="Regular", no_answer_number_of_rings: int=5, forward_timeout_seconds: int=0):
        """
        Builds a hunt group (HG) in the specified group. 

        Args:
            service_provider_id (str): The service provider ID in which the target group is built.
            group_id (str): The group ID where the HG should be built.
            service_user_id (str): The service user ID for the new HG. This must include the domain of the user.
            clid_first_name (str): The Calling Line ID first name.
            clid_last_name (str): The Calling Line ID last name. 
            extension (str): The extension number for the HG. This must be entered as a string. 
            payload (dict, optional): HG configuration data. 
            agents (list, optional): List of user IDs (str) that should be assigned to the new HG. The user(s) must already exist in the group. 
            policy (str, optional): Regular, Circular, Simultaneous, Uniform, Weighted. Defaults to Regular.
            no_answer_number_of_rings (int, optional): Defaults to 5.
            forward_timeout_seconds (int, optional): Defaults to 0.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/hunt-groups"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id 
        payload["serviceUserId"] = service_user_id
        payload["policy"] = policy
        payload["noAnswerNumberOfRings"] = no_answer_number_of_rings
        payload["forwardTimeoutSeconds"] = forward_timeout_seconds
        
        if agents:
            payload["agents"] = [{"userId": agent} for agent in agents]

        if 'serviceInstanceProfile' not in payload:     
            payload.setdefault('serviceInstanceProfile', {})            

        payload["serviceInstanceProfile"]["callingLineIdFirstName"] = clid_first_name
        payload["serviceInstanceProfile"]["callingLineIdLastName"] = clid_last_name
        payload["serviceInstanceProfile"]["name"] = f"{clid_first_name} {clid_last_name}"
        payload["serviceInstanceProfile"]["extension"] = extension

        return self.requester.post(endpoint, data=payload)
    

    def group_hunt_groups_remove_user(self, service_provider_id: str, group_id: str, user_id: str):

        """Removes the specified user from all hunt groups in which it currently exists. 

        Args:
            service_provider_id (str): The service provider ID in which the target user exists.
            group_id (str): The group ID where the user exists.
            user_id (str): The User ID of the user that is to be removed from the hunt group(s).

        Returns:
            List: Service user ID's (str) of the hunt groups from which the specified user has been removed. 
        """

        endpoint = "/groups/hunt-groups/removeUser"

        data = {
            "serviceProviderId": service_provider_id, 
            "groupId": group_id,
            "userId": user_id
        }

        return self.requester.post(endpoint, data=data)

#IN CALL SERVICE ACTIVATION
#INSTANT GROUP CALL
#INTEGRATED IMP
#INTERCEPT
#INTERNAL CALLING LINE ID DELIVERY
#LANGUAGES
#LEGACY AUTOMATIC CALLBACK
#MALICIOUS CALL TRACE
#MEDIA
#MEET-ME CONFERENCING
#MUSIC ON HOLD
#MWI DELIVERY TO MOBILE ENDPOINT
#NETWORK CLASS OF SERVICE
#NIGHT FORWARDING
#NUMBER PORTABILITY ANNOUNCEMENT
#NUMBER PORTABILITY QUERY
#NUMBERS
#OUTLOOK INTEGRATION
#PASSCODE RULES
#PASSWORD GENERATE
#PASSWORD RULES
#PERSONAL PHONE LIST
#PHONE DIRECTORY
#PHYSICAL LOCATION
#POLYCOM PHONE SERVICES
#PRE ALERTING
#PREFERRED CARRIER
#PREPAID
#PRIMARY ENDPOINT ADVANCED SETTING
#PRIORITY ALERT
#PRIVACY
#PUSH REGISTRATION
#PUSH TO TALK
#REGISTRATION
#REMOTE OFFICE
#REPORTS
#RESELLERS
#ROUTE LIST
#ROUTING PROFILE
#SCHEDULES
#SECURITY CLASSIFICATION
#SELECTIVE CALL ACCEPTANCE
#SELECTIVE CALL REJECTION
#SEQUENTIAL RING
#SERIES COMPLETION
#SERVICE PACKS
#SERVICE PROVIDERS
#SERVICES
#SHARED CALL APPEARANCE

    def user_shared_call_appearance_endpoint(self, user_id: str, line_port: str, device_name):
        """Creates a new Shared Call Apprance (SCA) on a single user.

        Args:
            user_id (str): Target user id of user to create SCA on.
            line_port (str): Line port to be assigned to the new SCA.
            device_name (_type_): Device to add for SCA from available devices.

        Returns:
            dict: New SCA details applied to user. 
        """
        
        endpoint = "/users/shared-call-appearance/endpoints"
        
        data = {
            "userId":user_id,
            "linePort":line_port,
            "isActive":True,
            "allowOrigination":True,
            "allowTermination":True,
            "deviceName":device_name,
            "deviceLevel":"Group"
        }
        
        return self.requester.post(endpoint, data=data)

#SILENT ALERTING
#SIMULTANEOUS RING PERSONAL
#SMARTY ADDRESS
#SPEED DIAL 100
#SPEED DIAL 8
#STATES AND PROVINCES
#SYSTEM
#TERMINATING ALTERNATE TRUNK IDENTITY
#THIRD PARTY VOICE MAIL SUPPORT
#THIRD PARTY EMERGENCY CALLING
#TIME ZONES
#TRUNK GROUPS

    def group_trunk_group(self, service_provider_id: str, group_id: str, trunk_name: str, max_active_calls: int, payload: dict={}, 
                          sip_authentication_username: str="", sip_authentication_password: str=""):
        """
        Builds a Trunk Group (TG) in the specified group. 

        Args:
            service_provider_id (str): The service provider ID in which the target group is built.
            group_id (str): The group ID where the HG should be built.
            trunk_name (str): The name of the new TG.
            max_active_calls (str): The maximum active calls to be set on the TG.
            payload (dict, optional): Configuration for the TG.
            sip_authentication_username (str, optional): The SIP authentication username for the TG. This field is required if "requireAuthentication" is set to "true". 
            sip_authentication_password (str, optional): The SIP authentication password for the TG. You can generate a password for this using get.sip_password_generator. This field is required if "requireAuthentication" is set to "true". 

        Note:
            Several fields are set to have default values. Please refer to the online documentation. 

        Returns:
            Dict: Returns the Trunk Group profile. 
        """

        endpoint = "/groups/trunk-groups"

        payload["name"] = trunk_name
        payload["maxActiveCalls"] = max_active_calls
        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id

        default_payload_values = {
            "capacityExceededTrapInitialCalls": 0, 
            "capacityExceededTrapOffsetCalls": 0,
            "clidSourceForScreenedCallsPolicy": "Profile Name Profile Number",
            "continuousOptionsSendingIntervalSeconds": 30,
            "failureOptionsSendingIntervalSeconds": 10,
            "failureThresholdCounter": 1,
            "invitationTimeout": 6,
            "inviteFailureThresholdCounter": 1,
            "inviteFailureThresholdWindowSeconds": 30,
            "pilotUserCallOptimizationPolicy": "Optimize For User Services",
            "pilotUserCallingLineAssertedIdentityPolicy": "Unscreened Originating Calls",
            "pilotUserCallingLineIdentityForEmergencyCallsPolicy": "No Calls",
            "pilotUserCallingLineIdentityForExternalCallsPolicy": "No Calls",
            "pilotUserChargeNumberPolicy": "No Calls",
            "successThresholdCounter": 1,
            "useSystemUserLookupPolicy": "true",
            "userLookupPolicy": "Basic",
            "requireAuthentication": "false"
            }
        for key, default_value in default_payload_values.items():
            payload.setdefault(key, default_value)

        if payload["requireAuthentication"]:
            payload["sipAuthenticationUserName"] = sip_authentication_username
            payload["sipAuthenticationPassword"] = sip_authentication_password

        return self.requester.post(endpoint, data=payload)

#TWO STAGE DIALING
#USERS
        
    def user(self, service_provider_id: str, group_id: str, user_id: str, first_name: str, last_name: str, 
             extension: str, web_auth_password: str, payload: dict={}):
        """
            Creates a new user in the specified group with the configuration defined in the payload.

        Args:
            service_provider_id (str): Service provider ID where Group is loctaed.
            group_id (str): Group ID where new user will be built.
            user_id (str): Complete User ID including group domain of new user.
            first_name (str): First name of new user.
            last_name (str): Last name of new user.
            extension (str): Extension number of new user.
            web_auth_password (str): Web authentication password. Note get.password_generate() can be used to get this.
            payload (dict, optional): User configuration.

        Returns:
            Dict: New user entity.
        """
        
        endpoint = "/users"
        
        payload["callingLineIdFirstName"] = first_name if not payload.get("callingLineIdFirstName") else payload["callingLineIdFirstName"]
        payload["callingLineIdLastName"] = last_name if not payload.get("callingLineIdLastName") else payload["callingLineIdLastName"]

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["userId"] = user_id
        payload["firstName"] = first_name
        payload["lastName"] = last_name
        payload["extension"] = extension
        payload["password"] = web_auth_password   
        
        return self.requester.post(endpoint, data=payload)
    
#USER CUSTOM RINGBACK
#VIDEO ADD ON
#VIRTUAL ON-NET ENTERPRISE EXTENSIONS
#VOICE MESSAGING
#VOICE PORTAL CALLING
#ZONE CALLING RESTRICTIONS
#ODIN BRANDING
#ODIN CALLBACKS
#ODIN CONNECTORS
#ODIN EMAIL
#ODIN EVENTS
#ODIN INVENTORY
#ODIN REPORTS
#ODIN SETTINGS
#ODIN STATUS
#ODIN SSO
#ODIN SSO ALTERNATE USER IDS
#ODIN TASKS
#ODIN UI
#ODIN VIEWABLE PACKS
#ODIN WEBHOOKS
#ODIN AUDIT
#ODIN IMPORTS
#ODIN EXPORTS
#XSI
#SIP AUTHENTICATION
#POLICY
#PASSWORD RESET
#PARTNERS
#USER UTILITIES
#ODIN TASKS COPY
#ODIN CONNECTORS
#LOCALES



from . import reports


class Reporter:
    """generates human friendly reports."""

    def __init__(self, api) -> None:
        self.api = api

    def call_flow(
        self,
        service_provider_id: str,
        group_id: str,
        number: str,
        number_type: str,
        broadworks_entity_type: str,
    ):
        """Generates a graphical flow chart of how a call to a specified number routes 
        through the broadworks system.

        Args:
            service_provider_id (str): Service Provider/ Enterprise where group is hosted.
            group_id (str): Group ID where target number for call flow is located. 
            number (str): Target number for call flow. 
            number_type (str): Type of number, options: "dn": Direct Number, "extension": Extension, "alias": Alias
            broadworks_entity_type (str): Broadworks entity type target number is associated with. \
            Options: "auto_attendant": Auto Attendant, "call_center": Call Center, "hunt_group": Hunt Group, "user": User
        """

        return reports.call_flow.main(
            self.api,
            service_provider_id,
            group_id,
            number,
            number_type,
            broadworks_entity_type,
        )

    def group_users_call_statistics(
        self,
        service_provider_id: str,
        group_id: str,
        start_date: str,
        end_date: str = None,
        start_time: str = "00:00:00",
        end_time: str = "23:59:59",
        time_zone: str = "Z",
    ):
        """Generates a CSV deatiling each users incoming and outgoing call statistics over 
        a specified period for a single group. Each row contains user extension, user ID, and call stats.

        Args:
            service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
            group_id (str): Target Group you would like to know user statistics for.
            start_date (str): Start date of desired time period. Date must follow format 'YYYY-MM-DD'
            end_date (str, optional): End date of desired time period. Date must follow format 'YYYY-MM-DD'.\
                If this date is the same as Start date you do not need this parameter. Defaults to None.
            start_time (_type_, optional): Start time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "00:00:00". MAX Request is 3 months.
            end_time (_type_, optional): End time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "23:59:59". MAX Request is 3 months.
            time_zone (str, optional): A specified time you would like to see call records in. \
                Time zone must follow format 'GMT', 'EST', 'PST'. Defaults to "Z" (UTC Time Zone).
        """
        return reports.group_users_call_statistics.main(
            self.api,
            service_provider_id,
            group_id,
            start_date,
            end_date,
            start_time,
            end_time,
            time_zone,
        )

    def user_registration_report(self, service_provider_id: str, group_id: str):
        """Generates an Excel Worksheet detailing each Users ID, device name and registration status within a group.

        Args:
            service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
            group_id (str): Target Group you would like to check the registration of.

        Returns:
            Xlsx File into .os_reports/ named "Registration_report_for_(GroupID)"
        """
        return reports.user_registration_report.main(
            self.api, service_provider_id, group_id
        )



import json

from tqdm import tqdm

from ..exceptions import OSServiceNotAssigned

def main(api: object, service_provider_id: str):
    """Returns a JSON breakdown of the Trunking Call Capacity of a Service Provider/ Enterprise (SP/ENT). 
    This will show the totals at each level from SP/ ENT to Group to Trunk Groups located in Groups. 
    At each level Max Active Calls and Bursting Max Active calls are detailed and then differences at 
    calculated.

    Raises: 
        OSServiceNotAssigned: Raises when the SP/ Ent does not have the service 'Trunk Group' assigned. 

    Args:
        service_provider_id (str): Target Service Provider ID/ Enterprise ID that you would like the \
            Trunk Call Capacity breakdown.

    Returns:
        JSON: JSON data of Trunking Call Capacity details of SP/ ENT, Groups, and Trunk Groups.
    """
    
    return_data = {}
    
    print(f"Fetching {service_provider_id} call capacity.")
    
    # OSServiceNotAssigned is returned if the Trunk Group service is not assigend to SP/ ENT
    try:
        service_provider_capacity = api.get.service_provider_trunk_group_call_capacity(service_provider_id)
    except Exception:
        raise OSServiceNotAssigned
        
    return_data["serviceProviderId"] = service_provider_capacity["serviceProviderId"]
    return_data["maxActiveCalls"] = service_provider_capacity["maxActiveCalls"]
    return_data["burstingMaxActiveCalls"] = service_provider_capacity["burstingMaxActiveCalls"]
    return_data["groupsCallCapacityTotal"] = 0
    return_data["groupsBurstingCallCapacityTotal"] = 0
    return_data["groups"] = []
    
    
    print(f"Fetching complete list of groups in {service_provider_id}.")
    groups_in_service_provider = api.get.groups(service_provider_id)
    
    # getting groups and group call capacities
    for group in tqdm(groups_in_service_provider, desc="Fetching groups call capacities."):
        formatted_group = {}
        formatted_group["groupId"] = group["groupId"]
        formatted_group["groupName"] = group["groupName"]
        
        try:
            group_capacity = api.get.group_trunk_groups_call_capacity(service_provider_id, group["groupId"])
            formatted_group["maxActiveCalls"] = group_capacity["maxActiveCalls"]
            formatted_group["burstingMaxAvailableActiveCalls"] = group_capacity["burstingMaxAvailableActiveCalls"]
            formatted_group["burstingMaxActiveCalls"] = group_capacity["burstingMaxActiveCalls"]
            formatted_group["burstingMaxAvailableActiveCalls"] = group_capacity["burstingMaxAvailableActiveCalls"]
            
            return_data["groupsCallCapacityTotal"] += group_capacity["maxActiveCalls"]
            return_data["groupsBurstingCallCapacityTotal"] += group_capacity["burstingMaxAvailableActiveCalls"]
               
        # group has no trunks and therefore fails
        except Exception:
        
            formatted_group["maxActiveCalls"] = 0
            formatted_group["burstingMaxAvailableActiveCalls"] = 0
            formatted_group["burstingMaxActiveCalls"] = 0
            formatted_group["burstingMaxAvailableActiveCalls"] = 0
            
            return_data["groupsCallCapacityTotal"] += 0
            return_data["groupsBurstingCallCapacityTotal"] += 0

        formatted_group["trunkGroups"] = []
        return_data["groups"].append(formatted_group)
    
    # getting trunk group capacities
    for group in tqdm(return_data["groups"], desc="Fetching trunk groups call capacities."):
        group["trunkGroupsCallCapacityTotal"] = 0
        group["trunkGroupsBurstingCallCapacityTotal"] = 0
        
        try:
            group_trunk_groups = api.get.group_trunk_groups(service_provider_id, group["groupId"])
            for trunk_group in group_trunk_groups:
                
                trunk_group_detailed = api.get.group_trunk_group(service_provider_id, group["groupId"], trunk_group["name"])
                
                group["trunkGroupsCallCapacityTotal"] += trunk_group_detailed["maxActiveCalls"]
                    
                group["trunkGroups"].append(
                    {
                        "name": trunk_group_detailed["name"],
                        "maxActiveCalls": trunk_group_detailed["maxActiveCalls"],
                        "burstingMaxActiveCalls": trunk_group_detailed["burstingMaxActiveCalls"] if "burstingMaxActiveCalls" in trunk_group_detailed else 0
                    }
                )
            
            group["callCapacityDifference"] = group["maxActiveCalls"] - group["trunkGroupsCallCapacityTotal"]
            group["burstingCallCapacityDifference"] = group["trunkGroupsBurstingCallCapacityTotal"] - group["trunkGroupsBurstingCallCapacityTotal"]
        
        # if no trunks exist in group api call fails
        except Exception:
            group["trunkGroupsCallCapacityTotal"] = 0
            group["trunkGroupsBurstingCallCapacityTotal"] = 0
            group["callCapacityDifference"] = 0
            group["burstingCallCapacityDifference"] = 0
        
    # working out totals difference 
    return_data["callCapacityDifference"] = return_data["maxActiveCalls"] - return_data["groupsCallCapacityTotal"]
    return_data["burstingCallCapacityDifference"] = return_data["burstingMaxActiveCalls"] - return_data["groupsBurstingCallCapacityTotal"]
        
    return return_data


import json

from dataclasses import dataclass, field
from typing import List, Type
   
@dataclass(kw_only=True)
class ServiceProvider:
    id: str
    name: str
    groups: List['Group'] = field(default_factory=list)
    is_enterprise: bool = False
    
    @classmethod
    def from_dict(cls, data):
        return cls(
            id= data.get("serviceProviderId"),
            name= data.get("serviceProviderId"),
            is_enterprise= data.get("isEnterprise")
        )
    
@dataclass(kw_only=True)
class Group:
    service_provider: Type['ServiceProvider']
    id: str
    name: str
    default_domain: str
    calling_line_id_phone_number: str = None
    auto_attendants: List['AutoAttendant'] = field(default_factory=list)
    trunk_groups: List['TrunkGroup'] = field(default_factory=list)
    call_centers: List['CallCenter'] = field(default_factory=list)
    hunt_groups: List['HuntGroup'] = field(default_factory=list)
    users: List['User'] = field(default_factory=list)

    def __post_init__(self):

        self.service_provider.groups.append(self)
        self.default_domain = '@' + self.default_domain
        
    @classmethod
    def from_dict(cls, service_provider: ServiceProvider, data):
        return cls(
            service_provider= service_provider,
            id= data.get("groupId"),
            name= data.get("groupName"),
            default_domain= data.get("defaultDomain"),
            calling_line_id_phone_number= data.get("callingLineIdPhoneNumber")
        )
        
@dataclass(kw_only=True)
class TrunkGroup:
    service_user_id: str
    group: Type['Group']
    users: List['User'] = field(default_factory=list)
    max_active_calls: int = None
    bursting_enabled: bool = False
    bursting_max_active_calls: bool = False
    pilot_user_id: str = None

    def __post_init__(self):
        self.group.trunk_groups.append(self)
        
    @classmethod
    def from_dict(cls, group: Group, data):
        
        # gather user IDs to gather user object
        user_ids = [agent["userId"] for agent in data["agents"]]
        users = _get_user_object_from_id(group, user_ids)
        
        return cls(
            service_user_id= data.get(""),
            group= group,
            users= users,
            max_active_calls= data.get("maxActiveCalls"),
            bursting_enabled= data.get("enableBursting"),
            bursting_max_active_calls= data.get("burstingMaxActiveCalls"),
            pilot_user_id= data.get("pilotUserId")
        )


@dataclass(kw_only=True)
class AAKey:
    number: int
    action: str
    description: str = None
    phone_number: str = None
    submenu_id: int = None
    
    @classmethod
    def from_dict(cls, data):
        return cls(
            number= data.get("key"),
            action= data.get("action"),
            description= data.get("description"),
            phone_number= data.get("phoneNumber"),
            submenu_id= data.get("submenuId")
        )


@dataclass(kw_only=True)
class AAMenu:
    enable_first_menu_level_extension_dialing: bool = False
    keys: List[AAKey] = field(default_factory=list)


@dataclass(kw_only=True)
class AutoAttendant:
    service_user_id: str
    name: str
    group: Type['Group']
    extension: str = None
    phone_number: str = None
    aliases: List[str] = field(default_factory= [list])
    type: str = None
    business_hours_menu: Type['AAMenu'] = None
    after_hours_menu: Type['AAMenu'] = None
    

    def __post_init__(self):
        self.group.auto_attendants.append(self)
        
        
    @classmethod
    def from_dict(cls, group: Group, data):
        return cls(
            service_user_id=data.get("serviceUserId"),
            name=data.get("serviceInstanceProfile").get("name"),
            group= group,
            extension=data.get("serviceInstanceProfile").get("extension"),
            phone_number=data.get("serviceInstanceProfile").get("phoneNumber"),
            aliases=data.get("serviceInstanceProfile").get("aliases"),
            type=data.get("type"),
            business_hours_menu=AAMenu(
                enable_first_menu_level_extension_dialing=data.get('businessHoursMenu').get('enableFirstMenuLevelExtensionDialing'),
                keys= [AAKey.from_dict(key) for key in data.get('businessHoursMenu').get('keys')]),
            after_hours_menu=AAMenu(
                enable_first_menu_level_extension_dialing=data.get('afterHoursMenu').get('enableFirstMenuLevelExtensionDialing'),
                keys= [AAKey.from_dict(key) for key in data.get('afterHoursMenu').get('keys')]),
        )

@dataclass(kw_only=True)
class CallCenter:
    service_user_id: str
    group: Type['Group']
    agents: List['User'] = field(default_factory=list)
    extension: str = None
    phone_number: str = None
    aliases: List[str] = field(default_factory=list)
    name: str = None
    type: str = None
    policy: str = None
   
    bounced_calls_enabled: bool = False
    overflow_calls_action: str = None
    overflow_calls_transfer_to_phone_number: bool = False
    stranded_calls_action: str = None
    stranded_calls_transfer_to_phone_number: bool = False
    stranded_call_unavailable_action: str = None
    stranded_call_unavailable_transfer_to_phone_number: bool = False
    
    #NOTE: Not sure which forwarding this is.
    forced_forwarding_enabled: bool = False
    forced_forwarding_forward_to_phone_number: str = None
    
    night_service: str = None
    holiday_service: str = None
    
    def __post_init__(self):
        self.group.call_centers.append(self)


    @classmethod
    def from_dict(cls, group: Group, data):
        
        # gather user IDs to gather user object
        try:
            agent_ids = [agent["userId"] for agent in data["agents"]]
            agents = _get_user_object_from_id(group, agent_ids)
        except KeyError: 
            agents = []
        
        return cls(
            service_user_id= data.get("serviceUserId"),
            group= group,
            agents= agents,
            extension= data.get("serviceInstanceProfile").get("extension"),
            phone_number=data.get("serviceInstanceProfile").get("phoneNumber"),
            name= data.get("serviceInstanceProfile").get("name"),
            aliases= data.get("serviceInstanceProfile").get("aliases"),
            type= data.get("type"),
            policy= data.get("policy"),
        
            bounced_calls_enabled= data.get("bouncedCallsEnabled"),
            overflow_calls_action= data.get("overFlowCallsAction"),
            overflow_calls_transfer_to_phone_number= data.get("overflowCallsTransferToPhoneNumber"),
            stranded_calls_action= data.get("strandedCallsAction"),
            stranded_calls_transfer_to_phone_number= data.get("strandedCallsTransferToPhoneNumber"),
            stranded_call_unavailable_action= data.get("strandedCallUnavailableAction"),
            stranded_call_unavailable_transfer_to_phone_number= data.get("strandedCallUnavailableTransferToPhoneNumber"),
            
            #NOTE: Not sure which forwarding this is.
            forced_forwarding_enabled= data.get("forcedForwardingEnabled"),
            forced_forwarding_forward_to_phone_number= data.get("forcedForwardingEnabled"),
            
            night_service= data.get("nightService"),
            holiday_service= data.get("holidayService")
        )

@dataclass(kw_only=True)
class HuntGroup:
    service_user_id: str
    name: str
    group: Type['Group']
    agents: List['User'] = field(default_factory=list)
    aliases: List[str] = field(default_factory=list)
    extension: str = None
    phone_number: str = None
    policy: str = None
    
    forward_after_timeout_enabled: bool = False
    forward_timeout_seconds: int = None
    no_answer_number_of_rings: int = None
    no_answer_forward_to_phone_number: str = None
    
    call_forward_not_reachable_enabled: bool = False
    call_forward_not_reachable_transfer_to_phone_number: str = None
    
    def __post_init__(self):
        self.group.hunt_groups.append(self)
    
    @classmethod
    def from_dict(cls, group: Group, data):

        # gather user IDs to gather user object
        agent_ids = [agent["userId"] for agent in data["agents"]]
        agents = _get_user_object_from_id(group, agent_ids)
        
        return cls(
            service_user_id= data.get("serviceUserId"),
            name= data.get("serviceInstanceProfile").get("name"),
            group= group,
            agents = agents,
            aliases= data.get("serviceInstanceProfile").get("aliases"),
            extension= data.get("serviceInstanceProfile").get("extension"),
            phone_number= data.get("serviceInstanceProfile").get("phoneNumber"),
            policy= data.get("policy"),
            
            forward_after_timeout_enabled= data.get("forwardAfterTimeout"),
            forward_timeout_seconds= data.get("forwardTimeoutSeconds"),
            no_answer_number_of_rings= data.get("noAnswerNumberOfRings"),
            no_answer_forward_to_phone_number= data.get("forwardToPhoneNumber"),
            
            call_forward_not_reachable_enabled= data.get("enableNotReachableForwarding"),
            call_forward_not_reachable_transfer_to_phone_number= data.get("notReachableForwardToPhoneNumber")
        )
        
@dataclass(kw_only=True)
class User:
    group: Type['Group']
    id: str
    first_name: str = None
    last_name: str = None
    extension: str = None
    phone_number: str = None
    aliases: List[str] = field(default_factory=list)
    
    call_forwarding_always: str = None
    call_forwarding_busy: str = None
    call_forwarding_no_answer: str = None
    call_forwarding_not_reachable: str = None


    def __post_init__(self):
        self.group.users.append(self)
 
 
    @classmethod
    def from_dict(cls, group: Group, data):
        return cls(
            group= group,
            id= data.get("userId"),
            first_name= data.get("firstName"),
            last_name= data.get("lastName"),
            extension= data.get("extension"),
            phone_number= data.get("phoneNumber"),
            aliases= data.get("aliases")
        )
        
@dataclass(kw_only=True)
class Contact:
    name: str = None
    number: str = None
    email: str = None
    
    @classmethod
    def from_dict(cls, data):
        return cls(
            name= data.get("name"),
            number= data.get("number"),
            email= data.get("email"),
        )
    
@dataclass(kw_only=True)
class Address:
    address_line1: str
    address_line2: str
    city: str
    state_or_province: str
    zip_or_postal_code: str
    country: str

    @classmethod
    def from_dict(cls, data):
        return cls(
            address_line1= data.get("addressLine1"),
            address_line2= data.get("addressLine2"),
            city= data.get("city"),
            state_or_province= data.get("stateOrProvince"),
            zip_or_postal_code= data.get("zipOrPostalCode"),
            country= data.get("country")
        )

@dataclass(kw_only=True)
class Department:
    service_provider_id: str
    group_id: str
    name: str
    
    @classmethod
    def from_dict(cls, data):
        return cls(
            service_provider_id= data.get("serviceProviderId"),
            group_id= data.get("groupId"),
            name= data.get("name")
        )
    
def _get_user_object_from_id(group, user_ids: list):
    return list(filter(lambda user: any(user_id in user.id for user_id in user_ids),
                       group.users))


import json
from tqdm import tqdm

def main(api, service_provider_id, group_id):
    """
        A stripped down version of group audit focussing only on the service packs assigned within
        the group. This only shows the service packs assigned and total count of unlike group audit 
        which details the users this is assigned to.

        :param service_provider_id (str): Service Provider ID or Enterprise ID containing the Group ID.
        :param group_id (str): Group ID to generate the report for.

        :return str: A JSON formatted report of service packs assigned in the group.
    """    
    
    service_report = api.get.group_services(group_id, service_provider_id)

    assigned_service_pack_services = []

    for sps in tqdm(service_report["servicePackServices"], desc="Analysing Service Packs..."):
            if sps["usage"] > 0:
                del sps["authorized"]
                del sps["assigned"]
                del sps["limited"]
                del sps["allowed"]
                del sps["serviceName"]
                del sps["quantity"]
                del sps["alias"]

                assigned_service_pack_services.append(sps)
                
    return {"servicePackServices": assigned_service_pack_services}



from ..utils.formatting import format_filter_value
from ..exceptions import *


class Get:

    filters = [
        "macAddress",
        "lastName",
        "firstName",
        "dn",
        "emailAddress",
        "userId",
        "extension",
    ]

    def __init__(self, requester):
        self.requester = requester

    # SESSION

    def session(self):

        endpoint = "/auth/session"

        return self.requester.get(endpoint)

    # ACCOUNT AUTHORIZATION CODES
    # ADMINISTRATORS
    # ADVICE OF CHARGE
    # ALTERNATE NUMBERS

    def user_alternate_numbers(self, user_id: str):
        """Fetches a list of a user/ service such as Auto Attendant, Hunt Group, or Call Centres
        alternate numebrs.

        Args:
            user_id (str): Target user/ service_user_id

        Returns:
            Dict: List of all alternate numbers assigned to the user/ service.
        """

        endpoint = "/users/alternate-numbers"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    # ANSWER CONFIRMATION
    # ALTERNATE USER ID
    # ANNOUNCEMENTS
    # ANONYMOUS CALL REJECTION
    # ATTENDANT CONSOLE
    # AUTHENTICATION
    # AUTO ATTENDANTS

    def auto_attendants(self, service_provider_id: str, group_id: str):
        """Returns a complete list of all Auto Attendants in a single group.

        Args:
            service_provider_id (str): Service Provider where Group is hosted.
            group_id (str): Target Group where Auto Attendants are hosted.

        Returns:
            List: List of Auto Attendants with basic info on them.
        """

        endpoint = "/groups/auto-attendants"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    def auto_attendant(self, service_user_id: str):
        """Returns detailed information of a single Auto Attendant.

        Args:
            service_user_id (str): User ID of target Auto Attendant.

        Returns:
            dict: Detailed information of target Auto Attendant.
        """

        endpoint = "/groups/auto-attendants"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def auto_attendant_user(
        self, service_provider_id: str, group_id: str, user_id: str
    ):
        """Returns detailed information about all Auto Attendants (AA) built in the same group as the specified user.

        Args:
            service_provider_id (str): Service Provider ID of the group where the user is built.
            group_id (str): Group ID of the group where the user is built.
            user_id (str): User ID of the user being queried.

        Returns:
           List: Returns a list of the AAs built in the group.
        """

        endpoint = "/groups/auto-attendants/user"

        params = {
            "userId": user_id,
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
        }

        return self.requester.get(endpoint, params=params)

    def auto_attendant_submenus(self, service_user_id: str):
        """Returns a list of the submenus of the specified Auto Attendant (AA). Works with Standard AAs only, basic AAs do not have submenus.

        Args:
            service_user_id (str): The service user ID of the AA

        Returns:
            List: Returns a list of the submenus associated with the AA.
        """

        endpoint = "/groups/auto-attendants/submenus"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def auto_attendant_submenu_usage(self, service_user_id: str, submenu_id: str):
        """Returns the type of the specified Auto Attendant (AA) submenu. NOTE: This method does not return any usage data.

        Args:
            service_user_id (str): The service user ID of the AA being queried.
            submenu_id (str): The submenu ID of the submenu being queried.

        Returns:
            List: Returns a list containing a single dict of the submenu.
        """

        endpoint = "/groups/auto-attendants/submenus/usage"

        params = {"serviceUserId": service_user_id, "submenuId": submenu_id}

        return self.requester.get(endpoint, params=params)

    # AUTOMATIC CALLBACK
    # AUTOMATIC HOLD RETRIEVE
    # BARGE IN EXEMPT
    # BASIC CALL LOGS
    # BROADWORKS ANYWHERE
    # BROADWORKS MOBILITY
    # BROADWORKS NAVIGATION
    # BROADWORKS RECEPTIONIST ENTERPRISE
    # BROADWORKS RECEPTIONIST OFFICE
    # BROADWORKS RECEPTIONIST SMALL BUSINESS
    # BUSY LAMP FIELD
    # CALL CAPACITY
    # CALL CENTER

    def group_call_centers(self, service_provider_id: str, group_id: str):
        """Retrieves a list of active call centers within a specified group, along with their settings.

        Args:
            service_provider_id (str): Target Service Provider where group is hosted
            group_id (str): Target Group ID

        Returns:
            List: List of Call Centers and their settings.
        """

        endpoint = "/groups/call-centers"
        params = {"serviceProviderId": service_provider_id, "groupId": group_id}
        return self.requester.get(endpoint, params=params)

    def group_call_center(self, service_user_id: str):
        """Retrieves deatiled information on a single Call Center.

        Args:
            service_user_id (str): Target Call Center's ID

        Returns:
            Dict: Target Call Centers details.
        """

        endpoint = "/groups/call-centers"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def user_call_center(self, user_id: str):
        """Retrieves a list of call centers that the specified user is currently associated with.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Agents Call Centers setting and a list of the User's associated Call Centers.
        """

        endpoint = "/users/call-center"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def group_call_center_agents(self, service_user_id: str):
        """Returns list of agents assigned to the target call center.

        Args:
            service_user_id (str): Service user ID of target call center.

        Returns:
            Dict: List of agents assigned to call center
        """

        endpoint = "/groups/call-centers/agents"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def group_call_center_agents(self, service_user_id: str):
        """Returns list of agents assigned to the target call center.

        Args:
            service_user_id (str): Service user ID of target call center.

        Returns:
            Dict: List of agents assigned to call center
        """

        endpoint = "/groups/call-centers/agents"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def group_call_center_bounced_calls(self, service_user_id: str):
        """Retrieves the number of rings before a call is bounced from the specified call center.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Amount of Rings before a call is Bounced
        """

        endpoint = "/groups/call-centers/bounced-calls"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def group_call_center_forced_forwarding(self, service_user_id: str):
        """Retrieves the forwarding number for a call center if it is set to forward calls, along with any associated audio messages.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Number to be Forwarded to, alongside any Audio Messages.
        """

        endpoint = "/groups/call-centers/forced-forwarding"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def group_call_center_overflow(self, service_user_id):
        """Retrieves the forwarding number for a user when all call center agents are busy, along with any associated audio messages.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Dict: Call Centers overflow configuration.
        """

        endpoint = "/groups/call-centers/overflow"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def group_call_center_stranded_calls(self, service_user_id):
        """Retrieves the forwarding number for a user when a call center doesn't answer, along with any associated audio messages.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Call Centers stranded call configuration.
        """

        endpoint = "/groups/call-centers/stranded-calls"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def group_call_center_stranded_calls_unavailable(self, service_user_id):
        """Retrieves the forwarding number for a user when a call center doesn't answer, along with the count of agents with an unavailable code in the call center.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Number to be Forwarded to, and Agents with an Unavailable Code set.
        """

        endpoint = "/groups/call-centers/stranded-calls-unavailable"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    # CALL CONTROL
    # CALL FORWARDING ALWAYS

    def user_call_forwarding_always(self, user_id: str):
        """Retrieves the Forwarding Always status for the specified User.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status, and the Number to be Forwarded to.
        """

        endpoint = "/users/call-forwarding-always"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def bulk_call_forwarding_always(self, service_provider_id: str, group_id: str):
        """Retrieves the Forwarding Always status for all users within a specified group.

        Args:
            service_provider_id (str): Target Service Provider where group is hosted
            group_id (str): Target Group ID

        Returns:
            List: Forwarding enabled status, the Number to be Forwarded to, and User information.
        """

        endpoint = "/users/call-forwarding-always/bulk"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    # CALL FORWARDING BUSY

    def user_call_forwarding_busy(self, user_id: str):
        """Retrieves the Forwarding Not Reachable status for the specified user.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status, and the Number to be Forwarded to.
        """

        endpoint = "/users/call-forwarding-busy"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def bulk_call_forwarding_busy(self, service_provider_id: str, group_id: str):
        """Retrieves the Forwarding Busy status for all users within a specified group.

        Args:
            service_provider_id (str): Target Service Provider where group is hosted
            group_id (str): Target Group ID


        Returns:
            List: Forwarding enabled status, the Number to be Forwarded to, and User information.
        """

        endpoint = "/users/call-forwarding-busy/bulk"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    # CALL FORWARDING NO ANSWER

    def user_call_forwarding_no_answer(self, user_id: str):
        """Retrieves the Forwarding No Answer status for the specified user

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status, and the Number to be Forwarded to.
        """

        endpoint = "/users/call-forwarding-no-answer"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def bulk_call_forwarding_no_answer(self, service_provider_id: str, group_id: str):
        """Retrieves the Forwarding No Answer status for all users within a specified group.

        Args:
            service_provider_id (str): Target Service Provider where group is hosted
            group_id (str): Target Group ID


        Returns:
            List: Forwarding enabled status, the Number to be Forwarded to, and User information.
        """

        endpoint = "/users/call-forwarding-no-answer/bulk"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    # CALL FORWARDING NOT REACHABLE

    def user_call_forwarding_not_reachable(self, user_id: str):
        """Retrieves the Forwarding Not Reachable status for the specified user

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status, and the Number to be Forwarded to.
        """

        endpoint = "/users/call-forwarding-not-reachable"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def bulk_call_forwarding_not_reachable(
        self, service_provider_id: str, group_id: str
    ):
        """Retrieves the Forwarding Not Reachable status for all users within a specified group.

        Args:
            service_provider_id (str): Target Service Provider where group is hosted
            group_id (str): Target Group ID


        Returns:
            List: Forwarding enabled status, the Number to be Forwarded to, and User information.
        """

        endpoint = "/users/call-forwarding-not-reachable/bulk"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    # CALL FORWARDING SELECTIVE

    def user_call_forwarding_selective(self, user_id: str):
        """Retrieves the Forwarding Selective status for a specified User, alongside the criteria.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status and the Forwarding criteria.
        """

        endpoint = "/users/call-forwarding-selective"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def user_call_forwarding_selective_criterias(self, user_id: str):
        """Retrieves the Forwarding Selective status for a specified User, alongside the criteria's assigned.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status and the Forwarding criteria's names and settings.
        """

        endpoint = "/users/call-forwarding-selective/criteria"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def user_call_forwarding_selective_criteria(self, user_id: str, criteria_name: str):
        """Retrieves the Forwarding Selective status for a specified User, alongside the specified Criteria

        Args:
            user_id (str): Target User ID
            criteria_name (str): Target Criteria Name

        Returns:
            Dict: Forwarding enabled status and the specified Criterias Settings.
        """

        endpoint = "/users/call-forwarding-selective/criteria"

        params = {"criteriaName": criteria_name, "userId": user_id}

        return self.requester.get(endpoint, params=params)

    # CALL NOTIFY
    # CALL PARK
    # CALL PICKUP

    def call_pickup_group_user(self, service_provider_id, group_id, user_id):
        """Retrieves Pickup Group information for the specified user.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): The Target Group ID the user is apart of.
            user_id (str): Target User ID

        Returns:
            Dict: Specified users pickup group, and the users within that group.
        """

        endpoint = "/groups/call-pickup/user"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "userId": user_id,
        }

        return self.requester.get(endpoint, params=params)

    # CALL POLICIES
    # CALL PROCESSING POLICIES
    # CALL RECORDING
    # CALL RECORDS

    def users_stats(
        self,
        user_id: str,
        start_date: str,
        end_date: str = None,
        start_time: str = "00:00:00",
        end_time: str = "23:59:59",
        time_zone: str = "Z",
    ):
        """Pulls a single users call statistics for a specified period of time. 

        Args:
            user_id (str): Target user ID you would like to pull call statistics for.
            start_date (str): Start date of desired time period. Date must follow format 'YYYY-MM-DD'
            end_date (str, optional): End date of desired time period. Date must follow format 'YYYY-MM-DD'.\
                If this date is the same as Start date you do not need this parameter. Defaults to None.
            start_time (_type_, optional): Start time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "00:00:00". MAX Request is 3 months.
            end_time (_type_, optional): End time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "23:59:59". MAX Request is 3 months.
            time_zone (str, optional): A specified time you would like to see call records in. \
                Time zone must follow format 'GMT', 'EST', 'PST'. Defaults to "Z" (UTC Time Zone).

        Returns:
            Dict: Users call record statistics for specified time period.
        """

        # checks if end_date has been left and therefore we assume user wants same date.
        if not end_date:
            end_date = start_date

        endpoint = "/users/call-records/stats"

        params = {
            "userIds": user_id,
            "startTime": f"{start_date}T{start_time}{time_zone}",
            "endTime": f"{end_date}T{end_time}{time_zone}",
        }

        return self.requester.get(endpoint, params=params)

    # CALL TRANSFER
    # CALL WAITING
    # CALLING LINE ID BLOCKING OVERRIDE
    # CALLING LINE ID DELIVERY BLOCKING
    # CALLING NAME DELIVERY
    # CALLING NAME RETRIEVAL
    # CALLING NUMBER DELIVERY
    # CALLING PARTY CATEGORY
    # CALLING PLANS
    # CALLBACKS
    # CHARGENUMBER
    # CLASSMARK
    # CLONE
    # COLLABORATE
    # COMM PILOT CALL MANAGER
    # COMM PILOT EXPRESS
    # COMMON PHONE LIST
    # COMMUNICATION BARRING
    # COMMUNICATION BARRING USER
    # CONNECTED LINE IDENTIFICATION
    # CUSTOM CONTACT DIRECTORY
    # DEPARTMENTS
    # DEVICE POLICIES
    # DEVICES
    # DIAL PLAN POLICY
    # DIRECTED CALL PICKUP WITH BARGE IN
    # DIRECTROUTE
    # DN

    def group_dns(self, service_provider_id: str, group_id: str):
        """Gets all numbers assigned to group.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            group_id (str): Group ID of target group.

        Returns:
            Dict: Dictionary containing all DNs assigned to group.
        """

        endpoint = "/groups/dns"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    def group_dn_search(
        self,
        service_provider_id: str,
        group_id: str,
        dn: int,
        filter_type: str = None,
        limit: int = None,
    ):
        """Searches for numbers assigned to group and allows search criteria and limiting result.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            group_id (str): Group ID of target group where numbers are located.
            dn (int): Number/ part of number to search for.
            filter_type (str, optional): Options: equal to, starts with, or contains. Defaults to None.
            limit (int, optional): Limits the amount of values API returns. Defaults to None.

        Returns:
            List: List of numbers matching search criteria
        """

        endpoint = "/groups/dns/search"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        if filter_type:
            params["dn"] = format_filter_value(filter_type, dn)
        if limit:
            params["limit"] = limit

        return self.requester.get(endpoint, params=params)

    def group_dn_details(self, service_provider_id: str, group_id: str):
        """Gets all numbers assigned to Group in detail. This will show where the number is assigned
        in a group such as which user or hunt group.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            group_id (str): Group ID of target group where numbers are located.

        Returns:
            Dict: Dictionary of numbers with details of each number.
        """

        endpoint = "/groups/dns/details"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    def system_dn_search(self, dn: int):
        """Searches for number from System level. This will return where the number is located on the system.
        It will show the Service Provider/ Enterprise, Group ID, and User ID the number is assigned to.

        Args:
            dn (int): Target number excluding country code e.g. 123456789

        Returns:
            List: List of dictionaries containing details of each number that fit the search criteria.
        """

        endpoint = "/system/dns/search"

        params = {"dn": f"+{dn}"}

        return self.requester.get(endpoint, params=params)

    def system_dn(self, dn: int):
        """Searches for number from System level. This will return where the number is located on the system.
        It will show the Service Provider/ Enterprise, Group ID, and User ID the number is assigned to.

        Args:
            dn (int): Target number excluding country code e.g. 123456789

        Returns:
            List: List of dictionaries containing details of each number that fit the search criteria.
        """

        endpoint = "/system/dns"

        params = {"phoneNumber": f"+{dn}"}

        return self.requester.get(endpoint, params=params)

    def system_dn_summary(self):
        """Returns all numbers assigned to system.

        Returns:
            List: List of all Service Providers/ Enterprises and numbers assigned in ranges.
        """

        endpoint = "/system/dns/summary"

        return self.requester.get(endpoint)

    def system_dn_utilization(self):
        """Returns DN statistics for each Service Provider/ Enterprise such as total DNs assigned.

        Returns:
            List: List of all Service Provider/ Enterprises with statistics of DNs for each.
        """

        endpoint = "/system/dns/utilization"

        return self.requester.get(endpoint)

    def service_provider_dn_search(
        self,
        service_provider_id: str,
        dn: int,
        filter_type: str = None,
        limit: int = None,
    ):
        """Searches for numbers assigned to Service Provider/ Enterprise and allows search criteria and limiting result.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            dn (int): Number/ part of number to search for.
            filter_type (str, optional): Options: equal to, starts with, or contains. Defaults to None.
            limit (int, optional): Limits the amount of values API returns. Defaults to None.

        Returns:
            List: List of numbers matching search criteria
        """

        endpoint = "/service-providers/dns/search"

        params = {"serviceProviderId": service_provider_id}

        if filter_type:
            params["dn"] = format_filter_value(filter_type, dn)
        if limit:
            params["limit"] = limit

        return self.requester.get(endpoint, params=params)

    def service_provider_dns(self, service_provider_id: str):
        """Returns all numbers assigned to Service Provider/ Enterprise with the group its assigned to
        and if the numbers can be deleted.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where target numbers are located.

        Returns:
            Dict: All numbers assigend to Service Provider/ Enterprise with group and delete status.
        """

        endpoint = "/service-providers/dns"

        params = {"serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    # DO NOT DISTURB

    def user_do_not_disturb(self, user_id: str):
        """Returns the specificied users DND and Ring Splash state.

        Args:
            user_id (str): Target user id to return data.

        Returns:
            Dict: States DND and Ring Splash status.
        """

        endpoint = "/users/do-not-disturb"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    # DOMAINS
    # EMERGENCY NOTIFICATIONS
    # EMERGENCY ZONES
    # ENTERPRISE TRUNKS
    # EXECUTIVE
    # EXECUTIVE ASSISTANT
    # EXTENSIONS
    # EXTERNAL CALLING LINE ID DELIVERY
    # EXTERNAL CUSTOM RINGBACK
    # FAX MESSAGING
    # FEATURE ACCESS CODES
    # FLEXIBLE SEATING
    # GROUP PAGING
    # GROUPS

    def groups(self, service_provider_id: str):
        """Returns the specificied Service Provider's Groups.

        Args:
            service_provider_id (str): Target Service Provider ID

        Returns:
            List: List of groups and their Names, alongside groupID's and userLimits.
        """

        endpoint = "/groups"

        params = {"serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    def group(self, service_provider_id, group_id):
        """Returns the specificied Group's settings and information.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID

        Returns:
            Dict: Returns information about the specified group, such as the DID, userCount and Domain.
        """

        endpoint = "/groups"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    # GROUP NAVIGATION
    # HOTELING GUEST
    # HOTELING HOST
    # HUNT GROUPS

    def group_hunt_groups(self, service_provider_id, group_id):
        """Returns a list of all the Hunt Groups within the specified Group.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID

        Returns:
            List: Returns a list of every Hunt Group within a Group, alongside their extension and name.
        """

        endpoint = "/groups/hunt-groups"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    def group_hunt_group(self, service_user_id):
        """Returns detailed information about the specified Hunt Group.

        Args:
            service_user_id (str): UserID of the target Hunt Group.

        Returns:
            Dict: Returns the specified Hunt Groups settings and information, such as group policies, agents, and extension.
        """

        endpoint = "/groups/hunt-groups"

        params = {"serviceUserId": service_user_id}

        return self.requester.get(endpoint, params=params)

    def group_hunt_group_user(self, service_provider_id, group_id, user_id):
        """Returns the Hunt Group's the specified User is apart of.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID
            user_id (str): Target User ID

        Returns:
            List: Returns the Hunt Group's a user is within, alongside their settings within that Hunt Group.
        """

        endpoint = "/groups/hunt-groups/user"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "userId": user_id,
        }

        return self.requester.get(endpoint, params=params)

    def group_hunt_groups_available_users(
        self, service_provider_id: str, group_id: str
    ):
        """Returns a list of all users within the service provider that are available to be assigned to a hunt group in the specified group.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID

        Returns:
            List: available users (dict)
        """

        endpoint = "/groups/hunt-groups/users"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    # IN CALL SERVICE ACTIVATION
    # INSTANT GROUP CALL
    # INTEGRATED IMP
    # INTERCEPT
    # INTERNAL CALLING LINE ID DELIVERY
    # LANGUAGES
    # LEGACY AUTOMATIC CALLBACK
    # MALICIOUS CALL TRACE
    # MEDIA
    # MEET-ME CONFERENCING
    # MUSIC ON HOLD
    # MWI DELIVERY TO MOBILE ENDPOINT
    # NETWORK CLASS OF SERVICE
    # NIGHT FORWARDING
    # NUMBER PORTABILITY ANNOUNCEMENT
    # NUMBER PORTABILITY QUERY
    # NUMBERS
    # OUTLOOK INTEGRATION
    # PASSCODE RULES
    # PASSWORD GENERATE

    def password_generate(self, service_provider_id: str, group_id: str) -> dict:
        """Generates a single passwords following the groups rules.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate password for.

        Returns:
            dict: Single password generated according to the groups rules.
        """

        endpoint = "/password/generate"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    def passwords_generate(
        self, service_provider_id: str, group_id: str, limit: int = 10
    ) -> dict:
        """Generates a multiple passwords to the limit set in pararmeters.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate password for.
            limit (int, optional): Number of passwords api will return. Defaults to 10.

        Returns:
            dict: Multiple passwords generated according to the groups rules.
        """

        endpoint = "/password/generate"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "limit": limit,
        }

        return self.requester.get(endpoint, params=params)

    def passcode_generate(self, service_provider_id: str, group_id: str) -> dict:
        """Generates a single passcode following group rules.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate passcode for.

        Returns:
            dict: Single passcode generated according to the groups rules.
        """

        endpoint = "/passcode/generate"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    def passcodes_generate(
        self, service_provider_id: str, group_id: str, limit: int = 10
    ) -> dict:
        """Generates a multiple passcodes to the limit set in pararmeters.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate passcodes for.
            limit (int, optional): Number of passwords api will return. Defaults to 10.

        Returns:
            dict: Multiple passcodes generated according to the groups rules.
        """

        endpoint = "/passcode/generate"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "limit": limit,
        }

        return self.requester.get(endpoint, params=params)

    def sip_password_generate(self) -> dict:
        """Generates a single SIP password.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate SIP password for.

        Returns:
            dict: Single SIP password generated according to the groups rules.
        """

        endpoint = "/sip-password/generate"

        return self.requester.get(endpoint)

    def sip_passwords_generate(self, limit: int = 10) -> dict:
        """Generates multiple SIP passwords to the limit set in parameters. Defaults to 10.

        Args:
            limit (int, optional): Number of passwords api will return. Defaults to 10.

        Returns:
            dict: Mutliple SIP passwords generated according to the groups rules.
        """

        endpoint = "/password/generate"

        params = {"limit": limit}

        return self.requester.get(endpoint, params=params)

    # PASSWORD RULES
    # PERSONAL PHONE LIST
    # PHONE DIRECTORY
    # PHYSICAL LOCATION
    # POLYCOM PHONE SERVICES
    # PRE ALERTING
    # PREFERRED CARRIER
    # PREPAID
    # PRIMARY ENDPOINT ADVANCED SETTING
    # PRIORITY ALERT
    # PRIVACY
    # PUSH REGISTRATION
    # PUSH TO TALK
    # REGISTRATION

    def user_registration(self, user_id: str):
        """Gets a users devices and if those devices are registered. This includes soft phones.

        Args:
            user_id (str): Target user ID to check registration

        Returns:
            dict: All users devices and details on device such as registration.
        """

        endpoint = "/users/registration"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def bulk_user_registration(self, service_provider_id: str, group_id: str):
        """Gets all users in a group and their device registrations. This includes soft phones.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is hosted.
            group_id (str): Target Group ID where users are located.

        Returns:
            dict: All users devices and details on device such as registration.
        """

        endpoint = "/users/registration/bulk"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    # REMOTE OFFICE
    # REPORTS

    def user_report(self, user_id: str):
        """Detailed report of user including services and service packs assigned.

        Args:
            user_id (str): Target user id of user.

        Returns:
            dict: Detailed report of user including services and service packs.
        """

        endpoint = "/users/reports/users"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    # RESELLERS
    # ROUTE LIST
    # ROUTING PROFILE
    # SCHEDULES

    def group_schedules(self, service_provider_id: str, group_id: str):
        """Retrieves the Business Schedules for the specified group.

        Args:
            service_provider_id (str): Target Service Provider ID where group is hosted.
            group_id (str): Target Group ID with schedules.

        Returns:
            List: List of all the groups schedules, including Name, Type and Level.
        """

        endpoint = "/groups/schedules"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self.requester.get(endpoint, params=params)

    def group_events(
        self, service_provider_id: str, group_id: str, name: str, type: str
    ):
        """Retrieves the Business Schedule's Events for the specified group.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID
            name (str): Name of the target Busisness Schedule
            type (str): The type of the Business Schedule (Time, Holiday)

        Returns:
            List: List of each Business Schedule Event, including startTime and endTime.
        """

        endpoint = "/groups/events"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "name": name,
            "type": type,
        }

        return self.requester.get(endpoint, params=params)

    # SECURITY CLASSIFICATION
    # SELECTIVE CALL ACCEPTANCE
    # SELECTIVE CALL REJECTION
    # SEQUENTIAL RING
    # SERIES COMPLETION
    # SERVICE PACKS
    # SERVICE PROVIDERS

    def service_providers(self, reseller_id=None):
        """
        Args:
            reseller_id (str): Only list the Service Provider IDs within the specified Reseller.
        """
        endpoint = "/service-providers"
        params = {"resellerId": reseller_id}

        return self.requester.get(endpoint, params=params)

    def service_provider(self, service_provider_id: str):
        """
        Args:
            reseller_id (str): Only list the Service Provider IDs within the specified Reseller.
        """

        endpoint = "/service-providers"

        params = {"serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    # SERVICES

    def user_services_assigned(self, user_id: str):
        """

        Args:
            user_id (str): _description_
        """
        endpoint = "/users/services/assigned"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def user_services(self, user_id: str):
        """Fetch all services assigned to a broadwrok entity, this can be
        a user, AA, CC, or HG.

        Args:
            user_id (str): User ID of the target Broadworks entity.

        Returns:
            Dict: Broadwork entiy and a list of services assigned.
        """

        endpoint = "/users/services"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def user_service_settings(self, user_id: str):
        """Retrieves all service settings for a specific user.

        Args:
            user_id (str): ID of the target user

        Returns:
            Dict: A dictionary containing all the service settings for the specified user.
        """

        endpoint = f"/users/services/settings"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)

    def group_services(self, group_id: str, service_provider_id: str):
        """
        Fetch all userServices, groupServices and servicePackServices assigned to a group.

        Args:
            group_id (str): GroupID of the target
            service_provider_id (str): Service Provider or Enterprise ID of the target.

        Returns:
            Dict: Authorised and assigned services within the group.
        """

        endpoint = "/groups/services"

        params = {"groupId": group_id, "serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    def group_services_user_assigned(
        self,
        group_id: str,
        service_provider_id: str,
        service_name: str,
        service_type: str,
    ):
        """
        Get details of the user/service instances where a particular service is assigned.

        Args:
            group_id (str): GroupID being targeted
            service_provider_id (str): Service provider/Enterprise ID where the group is located.
            service_name (str): Name of the service querying
            service_type (str): Type of service. Either: serviceName or servicePackName

        Returns:
            Dict: Users/Service Instances where the service is assigned.
        """

        endpoint = "/groups/services/assigned"

        params = {
            "groupId": group_id,
            "serviceProviderId": service_provider_id,
            "serviceType": service_type,
            "serviceName": service_name,
        }

        return self.requester.get(endpoint, params=params)

    # SHARED CALL APPEARANCE
    # SILENT ALERTING
    # SIMULTANEOUS RING PERSONAL
    # SMARTY ADDRESS
    # SPEED DIAL 100
    # SPEED DIAL 8
    # STATES AND PROVINCES
    # SYSTEM
    # TERMINATING ALTERNATE TRUNK IDENTITY
    # THIRD PARTY VOICE MAIL SUPPORT
    # THIRD PARTY EMERGENCY CALLING
    # TIME ZONES
    # TRUNK GROUPS

    def group_trunk_groups_call_capacity(self, service_provider_id: str, group_id: str):
        """Fetches Trunk Call Capacity data for a single Group.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located.
            group_id (str): Target Group to return data on.

        Returns:
            Dict: Trunk Group Call Capacity data of target group.
        """

        endpoint = "/groups/trunk-groups/call-capacity"

        params = {"groupId": group_id, "serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    def group_trunk_group(
        self, service_provider_id: str, group_id: str, trunk_group_name: str
    ):
        """Fetches all Trunk Group details of a single Trunk Group in a Group.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located.
            group_id (str): Group ID where the target Trunk Group is located.
            trunk_group_name (str): Target Trunk Group Name.

        Returns:
            Dict: Details of a target trunk group.
        """

        endpoint = "/groups/trunk-groups"

        params = {
            "groupId": group_id,
            "serviceProviderId": service_provider_id,
            "name": trunk_group_name,
        }

        return self.requester.get(endpoint, params=params)

    def group_trunk_groups(self, service_provider_id: str, group_id: str):
        """Fetches list of all trunk groups in a single group.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located.
            group_id (str): Group ID where the target Trunk Group is located.

        Returns:
            List: List of core details of all Trunk Groups located in a single Group.
        """

        endpoint = "/groups/trunk-groups"

        params = {"groupId": group_id, "serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    def service_provider_trunk_group_call_capacity(self, service_provider_id: str):
        """Fetches trunk call capacity details of a single Service Provider.

        Args:
            service_provider_id (str): Target Service Provider/ Enterprise ID.

        Returns:
            Dict: Trunk call capacity details of a single Service Provider/ Enterprise ID.
        """

        endpoint = "/service-providers/trunk-groups/call-capacity"

        params = {"serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    def service_provider_trunk_call_capacity_report(self, service_provider_id: str):
        """Fetches trunk call capacity details of Service Provider/ Enterprise and all Groups in the SP/ ENT.

        Args:
            servive_provider_id (str): Target Service Provider/ Enterprise ID.

        Returns:
            Dict: Breakdown of all trunk call capacity details of target Service Provider/ Enterprise and all Groups \
                in the target SP/ ENT.
        """

        endpoint = "/service-providers/trunk-groups/call-capacity/reports"

        params = {"serviceProviderId": service_provider_id}

        return self.requester.get(endpoint, params=params)

    # TWO STAGE DIALING
    # USERS

    def users(
        self,
        service_provider_id: str = None,
        group_id: str = None,
        filter: str = None,
        filter_type: str = None,
        filter_value: str = None,
        limit: int = None,
        extended=False,
    ):
        """
        Returns list of users depending on filter criteria.

        Args:
            servive_provider_id (str, optional): Service or Enterprise ID,
            top level object. Defaults to None.
            group_id (str, optional): Group ID where user is hosted. Defaults to None.
            filter (str, optional): Filter criteria, supported filters below. Defaults to None.
            filter_type (str, optional): Options: equal to, starts with, or contains. Defaults to None.
            filter_value (str, optional): Value filtering on e.g. firstName. Defaults to None.
            limit (int, optional): Limits the amount of values API returns. Defaults to None.

        Returns:
            dict: List of users.

        #Supported Filters
        macAddress: search by device
        lastName: filter by lastName
        firstName: filter by firstName
        dn: filter by dn
        emailAddress: filter by emailAddress
        userId: filter by userId
        extension: filter by extension

        ####Examples
        Get all users in Enterprise ent1
        GET /api/v2/users?serviceProviderId=ent1

        Get all users in Group grp1
        GET /api/v2/users?serviceProviderId=ent1&groupId=grp1

        Get up to 10 users in the system with a last name that contains smith
        GET /api/v2/users?lastName=*smith*&limit=10

        Get the users in grp1 that have a phone number that starts with 513333
        GET /api/v2/users?serviceProviderId=ent1&groupId=grp1&dn=513333*
        """

        endpoint = "/users?"

        params = {}

        if service_provider_id:
            params["serviceProviderId"] = service_provider_id
            if group_id:
                params["groupId"] = group_id
        if filter:
            params[filter] = format_filter_value(filter_type, filter_value)
        if limit:
            params["limit"] = limit
        if extended:
            params["extended"] = True

        return self.requester.get(endpoint, params=params)

    def user_by_id(self, user_id: str):
        """Returns extensive details of a single user including alias, enpoint device, and more common
        details like first and last name.

        Args:
            user_id (str): Target user ID of the user you would like to review.

        Returns:
            Dict: Python dictionary of the users details
        """

        endpoint = "/users"

        params = {"userId": user_id}

        return self.requester.get(endpoint, params=params)


# USER CUSTOM RINGBACK
# VIDEO ADD ON
# VIRTUAL ON-NET ENTERPRISE EXTENSIONS
# VOICE MESSAGING
# VOICE PORTAL CALLING
# ZONE CALLING RESTRICTIONS
# ODIN BRANDING
# ODIN CALLBACKS
# ODIN CONNECTORS
# ODIN EMAIL
# ODIN EVENTS
# ODIN INVENTORY
# ODIN REPORTS
# ODIN SETTINGS
# ODIN STATUS
# ODIN SSO
# ODIN SSO ALTERNATE USER IDS
# ODIN TASKS
# ODIN UI
# ODIN VIEWABLE PACKS
# ODIN WEBHOOKS
# ODIN AUDIT
# ODIN IMPORTS
# ODIN EXPORTS
# XSI
# SIP AUTHENTICATION
# POLICY
# PASSWORD RESET
# PARTNERS
# USER UTILITIES
# ODIN TASKS COPY
# ODIN CONNECTORS
# LOCALES



import json

from ..exceptions import OSInvalidPasswordType

def main(api, service_provider_id: str, group_id: str, users: list, password_type: str):
    
    # SIP auth password
    if password_type.lower() == "sip":
        
        print("Generating new SIP passwords.")
        new_passwords = api.get.sip_passwords_generate(len(users))['passwords']
        users_and_new_passwords = list(zip(users, new_passwords))
        
        print("Setting new SIP passwords.")
        for user in users_and_new_passwords:
            api.put.user_authentication_service(user[0], user[1])
        
        print("Setting new SIP passwords complete.")
        formatted_user_and_new_passwords = [{"userId": user[0], "newPassword": user[1]} for user in users_and_new_passwords]
        return formatted_user_and_new_passwords
    
    # intigration password
    elif password_type.lower() == "web":
        
        print("Generating new SIP passwords.")
        new_passwords = api.get.passwords_generate(service_provider_id, group_id, len(users))['passwords']
        users_and_new_passwords = list(zip(users, new_passwords))
        
        print("Setting new SIP passwords.")
        for user in users_and_new_passwords:
            api.put.user_web_authentication_password(user[0], user[1])
        
        print("Setting new SIP passwords complete.")
        formatted_user_and_new_passwords = [{"userId": user[0], "newPassword": user[1]} for user in users_and_new_passwords]
        return formatted_user_and_new_passwords
    
    # voicemail/ oortal
    elif password_type.lower() == "vm": # voicemail
        
        print("Generating new voicemail passcodes.")
        new_passcodes = api.get.passcodes_generate(service_provider_id, group_id, len(users))['passcodes']
        users_and_new_passcodes = list(zip(users, new_passcodes))
        
        print("Setting new voicemail passcodes.")
        for user in users_and_new_passcodes:
            api.put.user_portal_passcode(user[0], user[1])
          
        print("Setting new voicemail passcodes complete.")    
        formatted_user_and_new_passcodes = [{"userId": user[0], "newPasscode": user[1]} for user in users_and_new_passcodes]
        return formatted_user_and_new_passcodes
    
    # raise error if user attempts to change another type of password.
    else:
        raise OSInvalidPasswordType


import json
from tqdm import tqdm

def main(api, service_provider_id: str, group_id: str, user_id: str):
    """ identify a user's associations with Call Centers (CC), Hunt Groups (HG), 
    and Pick Up Groups.
    
    Args:
        service_provider_id (str): Service Provider where the group is hosted.
        group_id (str): Group where the User is located.
        user_id (str): Target user ID.
        
    Returns:
        str: Formatted output of the user showing all CC, HG, and Pick Up user is assigned to.
    """
    
    USER_DATA = {
        "userId": user_id,
        "firstName": None,
        "lastName": None,
        "extension": None,
        "phoneNumber": None,
        "aliases": None,
        "services": None,
        "featurePacks": None,
        "huntGroups": [],
        "callCenters": [],
        "pickUpGroup": None
    }
    
    try: 
        user = api.get.user_report(user_id)
    except Exception:
        return f"User {user_id} not found."
    
    USER_DATA["firstName"] = user["firstName"]
    USER_DATA["lastName"] = user["lastName"]
    USER_DATA["extension"] = user["extension"]
    USER_DATA["phoneNumber"] = user["phoneNumber"]
    USER_DATA["services"] = user["userServices"]
    USER_DATA["featurePacks"] = user["servicePacks"]
    USER_DATA["aliases"] = user["aliases"]
    
    pick_up_group = api.get.call_pickup_group_user(service_provider_id, group_id, user_id)
    try:
        USER_DATA["pickUpGroup"] = pick_up_group[0]["name"]
    except IndexError:
        USER_DATA["pickUpGroup"] = None
    
    hunt_groups = api.get.group_hunt_group_user(service_provider_id, group_id, user_id)
    for hg in tqdm(hunt_groups, desc="Fetching Hunt Groups"):
        USER_DATA["huntGroups"].append(hg["serviceUserId"])
    
    # if the user does not have a license for CC this call errors
    try:
        call_centers = api.get.user_call_center(user_id)
        for cc in tqdm(call_centers["callCenters"], desc="Fetching Call Centers"):
            USER_DATA["callCenters"].append(cc["serviceUserId"])
    except Exception:
        USER_DATA["callCenters"] = None
    
    return USER_DATA



import csv
import os

from tqdm import tqdm

from .report_utils.file_manager import copy_single_file_to_target_directory
from .report_utils.report_entities import call_records_statistics

def main(api: object, service_provider_id: str, group_id: str, 
         start_date:str, end_date: str = None, start_time: str = "00:00:00", 
         end_time:str = "23:59:59", time_zone: str = "Z"):
    """Generates a CSV deatiling each users incoming and outgoing call statistics over 
        a specified period for a single group. Each row contains user extension, user ID, and call stats.

    Args:
        service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
        group_id (str): Target Group you would like to know user statistics for.
        start_date (str): Start date of desired time period. Date must follow format 'YYYY-MM-DD'
        end_date (str, optional): End date of desired time period. Date must follow format 'YYYY-MM-DD'.\
            If this date is the same as Start date you do not need this parameter. Defaults to None.
        start_time (_type_, optional): Start time of desired time period. Time must follow formate 'HH:MM:SS'. \
            If you do not need to filter by time and want the whole day leave this parameter. Defaults to "00:00:00". MAX Request is 3 months.
        end_time (_type_, optional): End time of desired time period. Time must follow formate 'HH:MM:SS'. \
            If you do not need to filter by time and want the whole day leave this parameter. Defaults to "23:59:59". MAX Request is 3 months.
        time_zone (str, optional): A specified time you would like to see call records in. \
    """
    
    print("\nStart.")
    
    # List of report_entities.call_records_statistics
    group_users_statistics = []
    
    print(f"Fetching list of users in {group_id}.")
    
    # Fetches complete list of users in group
    users = api.get.users(service_provider_id, group_id)
    failed_users = []
    
    # Pulls stats for each user, instantiates call_records_statistics, and append to group_users_statistics
    for user in tqdm(users, "Fetching individual stats for each user. This may take several minutes"):
        
        try:
            user_statistics = api.get.users_stats(
                user["userId"],
                start_date,
                end_date,
                start_time,
                end_time,
                time_zone
            )
            
            user_services = api.get.user_services(
                user_id=user["userId"]
            )

        except Exception:
            # attempt 2 in case of connection time out
            try:
                user_statistics = api.get.users_stats(
                    user["userId"],
                    start_date,
                    end_date,
                    start_time,
                    end_time,
                    time_zone
                )
                
                user_services = api.get.user_services(
                user_id=user["userId"]
                )
                
            except Exception:
                failed_users.append(user)
                continue
        
        # Correction for API removing userId if no calls made by user
        if not user_statistics:
            user_statistics = {
                    "userId": user["userId"],
                    "total": 0,
                    "totalAnsweredAndMissed": 0,
                    "answeredTotal": 0,
                    "missedTotal": 0,
                    "busyTotal": 0,
                    "redirectTotal": 0,
                    "receivedTotal": 0,
                    "receivedMissed": 0,
                    "receivedAnswered": 0,
                    "placedTotal": 0,
                    "placedMissed": 0,
                    "placedAnswered": 0
                }
            
        user_statistics["servicePackServices"] = [service["serviceName"] for service in user_services["servicePackServices"] if service["assigned"]]
        
        user_statistic_record = call_records_statistics.from_dict(user["firstName"], user["lastName"], user["extension"], user_statistics)
        group_users_statistics.append(user_statistic_record)
    
    output_directory = "./os_reports"
    file_name = os.path.join(output_directory, f"{group_id} User Call Statistics - {start_date} to {end_date}.csv")

    # Ensure the directory exists
    os.makedirs(output_directory, exist_ok=True)   
    
    # Write statistics to csv 
    with open(file_name, mode="w", newline="") as file:
        fieldnames = [field.name for field in call_records_statistics.__dataclass_fields__.values()]
        writer = csv.DictWriter(file, fieldnames=fieldnames)
        writer.writeheader()
        
        for user in group_users_statistics:
            writer.writerow(user.__dict__)
        
        # Adds list of failed users to the bottom 
        if failed_users:
            writer.writerow({})
            writer.writerow({fieldnames[1]: "Failed Users"})
            
            for failed_user in failed_users:
                writer.writerow({fieldnames[1]: failed_user['userId']})
    
    # Add made_with_os.png for output   
    copy_single_file_to_target_directory("./odins_spear/assets/images/", "./os_reports/", "made_with_os.png")        
    print("\nEnd.")
            


from typing import List
import json

from .. import api
from ..utils import parsing
from . import broadwork_entities as bre

class DataStore:
    """ Local store of objects, when each object is instantiated it is added to
    the appropriate list.

    Singleton design pattern to get store use get_instance().
    """

    def __init__(self):
        self.apis: List[api.Api] = []
        self.service_providers_enterprises: List[bre.ServiceProvider] = []
        self.groups: List[bre.Group] = []
        self.trunk_groups: List[bre.TrunkGroup] = []
        self.auto_attendants: List[bre.AutoAttendant] = []
        self.call_centers: List[bre.CallCenter] = []
        self.hunt_groups: List[bre.HuntGroup] = []
        self.users: List[bre.User] = []
        self.devices: List[bre.Device] = []
        self.other_entities = [] #Non-common or custom objects


    def build_id_mapping(self):
        """
        Builds mapping of numbers IDs to entity.
        
        Example: {test@test.com: User, customID: CallCenter}
        """
        
        self.id_mapping = {}
        
        entities = self.auto_attendants + self.call_centers + \
        self.hunt_groups + self.users
        
        for e in entities:
            try:
                self.id_mapping[e.id] = e
            except Exception:
                self.id_mapping[e.service_user_id] = e
        
    
    def build_number_mapping(self):
        """
        Builds mapping of numbers (phone numbers, extension, aliases) to entity.
        
        Example: {101: User, +1-123456789: CallCenter}
        """
        import re
    
        self.number_mapping = {}
        
        entities = self.auto_attendants + self.call_centers + \
        self.hunt_groups + self.users
        
        for e in entities: 
            if e.phone_number:
                self.number_mapping[e.phone_number] = e
            if e.extension:
                self.number_mapping[e.extension] = e 
            if e.aliases:
                for a in e.aliases:
                    number = re.search(r'\d+', a).group()
                    self.number_mapping[number] = e     
        
      
    def get_group_state(api, group: bre.Group) -> None:
        """ takes in group id and loads group state into broadworks entities.

        :param api: API object used to send requests to create group state in store.
        :param group: Group object of Broadworks group user would like to load state.
        """
        pass

    def store_objects(self, *entities) -> None:
        """ Takes in objects within the odin_api and custom and stores in lists
        depending on type.

        :param entity: broadwork entities used in odin_api
        """

        for e in entities:
            if isinstance(e, api.Api):
                self.apis.append(e)
            elif isinstance(e, bre.ServiceProvider):
                self.service_providers_enterprises.append(e)
            elif isinstance(e, bre.Group):
                self.groups.append(e)
            elif isinstance(e, bre.TrunkGroup):
                self.trunk_groups.append(e)
            elif isinstance(e, bre.AutoAttendant):
                self.auto_attendants.append(e)
            elif isinstance(e, bre.CallCenter):
                self.call_centers.append(e)
            elif isinstance(e, bre.HuntGroup):
                self.hunt_groups.append(e)
            elif isinstance(e, bre.User):
                self.users.append(e)
            elif isinstance(e, bre.Device):
                self.devices.append(e)
            else:
                self.other_entities.append(e)
    
    def export_store(self) -> str:
        """Export all objects in the store and their relationships in JSON format."""
        export_data = {}
        object_lists = {
            'apis': self.apis,
            'service_providers': self.service_providers,
            'enterprises': self.enterprises,
            'groups': self.groups,
            'trunk_groups': self.trunk_groups,
            'auto_attendants': self.auto_attendants,
            'call_centers': self.call_centers,
            'hunt_groups': self.hunt_groups,
            'users': self.users,
            'devices': self.devices,
            'other_entities': self.other_entities
        }

        for key, object_list in object_lists.items():
            export_data[key] = parsing.export_objects(object_list)

        return json.dumps(export_data, indent=2)

    def __str__(self) -> str:
        """ returns complete list of entities in store.
        """
        entities = self.apis + self.enterprises + self.service_providers + self.groups + self.trunk_groups + \
            self.hunt_groups + self.users

        # loops entities and joins into string
        string = lambda e: "\n".join(map(str, e))
        return string(entities)



import re
from tqdm import tqdm

from ..exceptions import OSAliasNotFound


def locate_alias(alias, aliases: list):
    for a in aliases:
        if re.search(rf'\b{alias}\b', a):
            return True
        

def main(api, service_provider_id: str, group_id: str, alias: str):
    """ Locates alias if assigned to broadworks entity. 

    The script searches through various entity types including Auto Attendants (AA),
    Hunt Groups (HG), and Call Centers (CC), as well as individual Users. It employs 
    a retry mechanism for instances where initial attempts to fetch entity details fail.

    The search is conducted in two phases:
    1. Collecting details of AAs, HGs, and CCs and checking for the alias.
    2. If not found, searching through the users for the alias.

    If the alias is found, the method returns a string specifying the type of entity and
    its name or userID. If the alias is not found after checking all entities, an 
    AOAliasNotFound exception is raised.

    :param service_provider_id: Service Prodiver where group is hosted.
    :param group_id: Group where alias is located.
    :param alias: Alias number to identify e.g. 0

    :return str: Returns type and name/ userId of entity where alias located. 
    :raise AOALiasNotFound: If alias not found AOAliasNotFound error raised 
    """   

    RETRY_QUEUE = []
    MAX_RETRIES = 2  
    OBJECT_WITH_ALIAS = []
    
    auto_attendants = api.get.auto_attendants(service_provider_id, group_id)
    hunt_groups = api.get.group_hunt_groups(service_provider_id, group_id)
    call_centers = api.get.group_call_centers(service_provider_id, group_id)
    
    broadwork_entities_user_ids = []
    
    for aa in auto_attendants:
        broadwork_entities_user_ids.append(["AA", aa["serviceUserId"]])
        
    for hg in hunt_groups:
        broadwork_entities_user_ids.append(["HG", hg["serviceUserId"]])
        
    for cc in call_centers:
        broadwork_entities_user_ids.append(["CC", cc["serviceUserId"]])
        
    for broadwork_entity in tqdm(broadwork_entities_user_ids, desc="Fetching AA, HG, and CC details"):
        # add some buffer time for odins api 
        
        formatted = {
            "type":broadwork_entity[0],
            "service_user_id":broadwork_entity[1]
        }

        temp_object = ""
        
        try:
            if broadwork_entity[0] == "AA":
                temp_object = api.get.auto_attendant(broadwork_entity[1])
            elif broadwork_entity[0] == "HG":
                temp_object = api.get.group_hunt_group(broadwork_entity[1])
            else:
                temp_object = api.get.group_call_center(broadwork_entity[1])
                
            formatted["name"] = temp_object["serviceInstanceProfile"]["name"]
            formatted["aliases"] = temp_object["serviceInstanceProfile"]["aliases"]
            
            OBJECT_WITH_ALIAS.append(formatted)
            
        except Exception:
            # add a retry count and add this entity to retry queue
            broadwork_entity.append(0)
            RETRY_QUEUE.append(broadwork_entity)
             

    # objects failed in first instance
    if RETRY_QUEUE:
        print("Retrying failed instances.")
    while RETRY_QUEUE:
        entity_type, service_user_id, retry_count = RETRY_QUEUE.pop(0)  # Get the first item from the queue
        
        formatted = {}
        formatted["type"] = entity_type
        temp_object = ""
        
        try:
            if entity_type == "AA":
                temp_object = api.get.auto_attendant(service_user_id)
            elif entity_type == "HG":
                temp_object = api.get.group_hunt_group(service_user_id)
            else:
                temp_object = api.get.group_call_center(service_user_id)
                
            formatted["name"] = temp_object["serviceInstanceProfile"]["name"]
            formatted["aliases"] = temp_object["serviceInstanceProfile"]["aliases"]
            
            OBJECT_WITH_ALIAS.append(formatted)
            
        except Exception:
            if retry_count < MAX_RETRIES:
                RETRY_QUEUE.append((entity_type, service_user_id, retry_count + 1))  # Increment retry count and re-add to the queue
            else:
                print(f"Failed to process {entity_type} - {service_user_id} after {MAX_RETRIES} retries. Skipping.")

    for broadwork_entity in tqdm(OBJECT_WITH_ALIAS, desc=f"Searching AA, HG, and CC for alias {alias}"):

        if locate_alias(alias, broadwork_entity['aliases']):
            return broadwork_entity
        
    users = api.get.users(service_provider_id, group_id, extended=True)
    print("Fetched users.")
    
    for user in tqdm(users, desc=f"Searching Users for alias: {alias}"):
        
        if locate_alias(alias, user['aliases']):
            return {
                "type":  "user",
                "user_id": user['userId'],
                "alias": alias
            }
    
    return OSAliasNotFound


import json

def json_to_dictionary(json_data: str):
    """Loads JSON into a Python Dict.

    Args:
        json_data (str): JSON data in a string format.

    Returns:
        Dict: Python dictionary of JSON data.
    """
    try:
        return json.loads(json_data)
    except json.JSONDecodeError as e:
        print(f"JSON decoding error: {e}")
        return None
    
    
def dictionary_to_json(python_dict: dict):
    """Takes in a Python dictionary and converts to JSON.

    Args:
        python_dict (dict): Python dictionary.

    Returns:
        Str: JSON data in string format.
    """
    try:
        return json.dumps(python_dict)
    except Exception as e:
        # Handle any exceptions that may occur during JSON encoding
        print(f"Error converting dictionary to JSON: {str(e)}")
        return None


import pandas as pd


def export_to_xlsx(data: dict, group_id: str):
    rows = []
    print("Start.\n")
    print("Generating report.\n")

    for user_id, user_data in data.items():
        registration = user_data.get("registration", {})

        if not registration:
            rows.append([user_id, None, False])
        else:
            for device_name, registration_info in registration.items():
                rows.append(
                    [user_id, device_name, registration_info.get("registered", False)]
                )

    dataframe = pd.DataFrame(rows, columns=["User ID", "Device Name", "Registered"])
    dataframe.to_excel(
        f"./os_reports/Registration_report_for_{group_id}.xlsx",
        index=False,
    )
    print("End.\n")


def main(api, service_provider_id: str, group_id: str):
    """Generates an Excel Worksheet detailing each Users ID, device name and registration status within a group.

    Args:
        service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
        group_id (str): Target Group you would like to check the registration of.

    Returns:
            Xlsx File into .os_reports/ named "Registration_report_for_(GroupID)"
    """
    data = api.scripter.user_registration(service_provider_id, group_id)

    export_to_xlsx(data, group_id)



from tqdm import tqdm

from .report_utils.graphviz_module import GraphvizModule
from .report_utils.helpers import find_entity_with_number_type
from .report_utils.parsing import call_flow_module


from ..store import DataStore
from ..store import broadwork_entities as bre

def main(api, service_provider_id: str, group_id: str, number: str, number_type: str,
         broadworks_entity_type: str):
    
    # Creates data store for use later
    data_store = DataStore()
    
    print("Start.\n")
    print("Fetching Service Provider & Group details.")
    # Gather entities 
    service_provider = bre.ServiceProvider.from_dict(data=api.get.service_provider(service_provider_id))
    group = bre.Group.from_dict(service_provider=service_provider, data=api.get.group(service_provider_id, group_id))
    
    data_store.store_objects(service_provider, group)
    
    auto_attendants = api.get.auto_attendants(service_provider_id, group_id)
    for aa in tqdm(auto_attendants, desc=f"Fetching all Auto Attendants."):
        auto_attendant = bre.AutoAttendant.from_dict(group=group, data=api.get.auto_attendant(aa['serviceUserId']))
        data_store.auto_attendants.append(auto_attendant)
    
    print("Fetching all users this may take a couple of minutes, please wait.")
    users = api.get.users(service_provider_id, group_id, extended=True)
    
    # Captures users with the forward fucntionality
    call_forward_always_users = [
    item["user"]["userId"] for item in api.get.bulk_call_forwarding_always(service_provider_id, group_id) if item["service"]["assigned"] \
        and item["data"]["isActive"]
    ]

    call_forward_busy_users = [
    item["user"]["userId"] for item in api.get.bulk_call_forwarding_busy(service_provider_id, group_id) if item["service"]["assigned"] \
        and item["data"]["isActive"]
    ]
    
    call_forward_no_answer_users = [
    item["user"]["userId"] for item in api.get.bulk_call_forwarding_no_answer(service_provider_id, group_id) if item["service"]["assigned"] \
        and item["data"]["isActive"]
    ]
    
    call_forward_not_reachable = [
    item["user"]["userId"] for item in api.get.bulk_call_forwarding_not_reachable(service_provider_id, group_id) if item["service"]["assigned"] \
        and item["data"]["isActive"]
    ]

    for u in tqdm(users, desc=f"Parsing all Users."):
        user = bre.User.from_dict(group=group, data=u)
        
        if user.id in call_forward_always_users:
            user.call_forwarding_always = str(api.get.user_call_forwarding_always(user.id)["forwardToPhoneNumber"])
        if user.id in call_forward_busy_users:
            user.call_forwarding_busy = str(api.get.user_call_forwarding_busy(user.id)["forwardToPhoneNumber"])
        if user.id in call_forward_no_answer_users:
            user.call_forwarding_no_answer = str(api.get.user_call_forwarding_no_answer(user.id)["forwardToPhoneNumber"])
        if user.id in call_forward_not_reachable:
            user.call_forwarding_not_reachable = str(api.get.user_call_forwarding_not_reachable(user.id)["forwardToPhoneNumber"])
        
        data_store.users.append(user)


    call_centers = api.get.group_call_centers(service_provider_id, group_id)    
    for cc in tqdm(call_centers, desc=f"Fetching all Call Centers."):
        
        call_center = api.get.group_call_center(cc['serviceUserId'])
        call_center['agents'] = api.get.group_call_center_agents(cc['serviceUserId'])['agents']
        
        call_center = bre.CallCenter.from_dict(group= group, data= call_center)
        
        try:
            overflow_settings = api.get.group_call_center_overflow(call_center.service_user_id)
            call_center.overflow_calls_action = overflow_settings["action"]
            call_center.overflow_calls_transfer_to_phone_number = overflow_settings["transferPhoneNumber"] \
                if call_center.overflow_calls_action == "Transfer" else None
        except Exception:
            call_center.overflow_calls_action = None
            call_center.overflow_calls_transfer_to_phone_number = None
        
        try:
            stranded_calls_settings = api.get.group_call_center_stranded_calls(call_center.service_user_id)
            call_center.stranded_calls_action = stranded_calls_settings["action"]
            call_center.stranded_calls_transfer_to_phone_number = stranded_calls_settings["transferPhoneNumber"] \
                if call_center.stranded_calls_action == "Transfer" else None
        except Exception:
            call_center.stranded_calls_action = None
            call_center.stranded_calls_transfer_to_phone_number = None
            
        try:
            stranded_calls_unavailable_settings = api.get.group_call_center_stranded_calls_unavailable(call_center.service_user_id)
            action_value = stranded_calls_unavailable_settings["action"]
            call_center.stranded_call_unavailable_action = None if action_value == 'None' else action_value
            call_center.stranded_call_unavailable_transfer_to_phone_number = stranded_calls_unavailable_settings["transferPhoneNumber"] \
                if call_center.stranded_call_unavailable_action == "Transfer" else None
        except Exception:
            call_center.stranded_call_unavailable_action = None
            call_center.stranded_call_unavailable_transfer_to_phone_number = None
        
        try: 
            forced_forwarding_settings = api.get.group_call_center_forced_forwarding(call_center.service_user_id)
            call_center.forced_forwarding_enabled = forced_forwarding_settings["enabled"]
            call_center.forced_forwarding_forward_to_phone_number = forced_forwarding_settings["forwardToPhoneNumber"] \
                if call_center.forced_forwarding_enabled else None
        except Exception:
            call_center.forced_forwarding_enabled = False
            call_center.forced_forwarding_forward_to_phone_number = None
        
        data_store.call_centers.append(call_center)
    
    hunt_groups = api.get.group_hunt_groups(service_provider_id, group_id)
    for hg in tqdm(hunt_groups, desc=f"Fetching all Hunt Groups."):
        hunt_group = bre.HuntGroup.from_dict(group=group, data= api.get.group_hunt_group(hg['serviceUserId']))
        data_store.hunt_groups.append(hunt_group)
      
    # locate number using broadworks_entity_type to zone in on correct location
    call_flow_start_node = find_entity_with_number_type(
			number,
			number_type.lower(),
			getattr(data_store, broadworks_entity_type + "s")
		)
    call_flow_start_node._start_node = True
    
    # Nodes used in the graph
    print("Gathering nodes in flow.")
    bre_nodes = call_flow_module(call_flow_start_node, data_store)
    
    print("Generating report.")
    # build, generate, save graph
    graph = GraphvizModule(
        "./os_reports/"
    )
    graph.generate_call_flow_graph(
        bre_nodes,
        number
    )
    print("Saving report.")
    graph._save_graph(f"Calls To {number}")
    print("\nEnd.")


from dataclasses import dataclass, field, fields
from typing import List

@dataclass
class call_flow:
    name: str
    nodes: List = field(default_factory=list)


@dataclass
class external_number:
    id: str    
    
@dataclass
class call_records_statistics:
    first_name: str
    last_name: str
    extension: str
    userId: str
    feature_packs: list 
    total: int
    totalAnsweredAndMissed: str
    answeredTotal: str
    missedTotal: str
    busyTotal: str
    redirectTotal: str
    receivedTotal: str
    receivedMissed: str
    receivedAnswered: str
    placedTotal: str
    placedMissed: str
    placedAnswered: str

    @classmethod
    def from_dict(cls, first_name, last_name, extension, data):
        return cls(
            first_name = first_name,
            last_name  = last_name,
            extension = extension,
            userId= data.get("userId"),
            feature_packs = data.get("servicePackServices"),
            total= data.get("total"), 
            totalAnsweredAndMissed= str(data.get("totalAnsweredAndMissed")),
            answeredTotal= data.get("answeredTotal"),
            missedTotal= data.get("missedTotal"),
            busyTotal= data.get("busyTotal"),
            redirectTotal= data.get("redirectTotal"),
            receivedTotal= data.get("receivedTotal"),
            receivedMissed= data.get("receivedMissed"),
            receivedAnswered= data.get("receivedAnswered"),
            placedTotal= data.get("placedTotal"),
            placedMissed= data.get("placedMissed"),
            placedAnswered= data.get("placedAnswered")
        )



import unittest
import os
import sys


class TestOnApiInstantiation(unittest.TestCase):
    """ 
    Test library is setting up correctly when instatiating API object e.g.ensuring objects are 
    handed to other objects that need them to function. Finally, tests authentication needed
    to utilise the library.
    """
    
    
    def setUp(self):
        sys.path.append('..')
        
        from odins_spear.api import Api
        # Set up API object
        self.api = Api(base_url=os.getenv("ODIN_URL"), username= os.getenv("ODIN_USERNAME"), 
               password="ODIN_PASSWORD")
        

    def test_logger_set_up(self):
        """ Logger has been set up and username set correctly
        """
        self.assertEqual(
            self.api.username,
            self.api.logger._user
        )

    def test_requester_set_up(self):
        """ Logger has been set up and username set correctly
        """
        # base url
        self.assertEqual(
            self.api.base_url,
            self.api._requester.base_url
        )

        # rate limit
        self.assertEqual(
            self.api.rate_limit,
            self.api._requester.rate_limit
        )
        
        # logger
        self.assertIs(
            self.api.logger,
            self.api._requester.logger
        )

    
    def test_method_set_up(self):
        """ Methods are correctly setup with requester
        """
        # GET
        self.assertIs(
            self.api._requester,
            self.api.get._requester
        )

        # POST
        self.assertIs(
            self.api._requester,
            self.api.post._requester
        )
        
        # PUT
        self.assertIs(
            self.api._requester,
            self.api.put._requester
        )
        
        # DELETE
        self.assertIs(
            self.api._requester,
            self.api.delete._requester
        )
        
    def test_scripter_set_up(self):
        """ Scripter is setup with the API needed
        """
        self.assertIs(
            self.api.scripter.api,
            self.api
        )
        
        
    def test_reporter_set_up(self):
        """ Reporter is setup with the API needed
        """
        self.assertIs(
            self.api.reporter.api,
            self.api
        )


    def test_authentication(self):
        """ API object is able to authenticate.
        """
        try:
            self.assertEqual(
                self.api.authenticate(), 
                True
            )
        except Exception:
            self.fail("Authentication failed.")
            

if __name__ == '__main__':
    unittest.main()


import graphviz

from ...store import broadwork_entities as bre

class GraphvizModule:
    
    """
    Colours:
    - Black: 020300
    - Orange: FFA400 - Darker: B87700
    - Red: 95190C - Darker: 5E1008
    - Purple: 5B5F97 - Darker: 3D4066
    - Redish: A23E48 - Darker: 68272E
    """
    
    NODE_STYLING = {
        'start' : {
            'shape': 'record',
            'style': 'filled',
            'margin': '0.2',
            'color': '#020300',
            'fontname': 'Arial',
            'fontcolor': 'white'
        },
        'exit': {
            'shape': 'record',
            'style': 'filled',
            'margin': '0.2',
            'color': '#020300',
            'fontname': 'Arial',
            'fontcolor': 'white'
        },
        'auto_attendant': {
            'shape': 'record',
            'style': 'filled',
            'margin': '0.2',
            'color': '#1C3A3A',
            'fillcolor': '#356969',
            'fontname': 'Arial',
            'fontcolor': 'white'
        },
        'call_centre': {
            'shape': 'record',
            'style': 'filled',
            'margin': '0.2',
            'color': '#CCCCCC',
            'fillcolor': '#D9D9D9',
            'fontname': 'Arial',
            'fontcolor': 'black'
        },
        'hunt_group': {
            'shape': 'record',
            'style': 'filled',
            'margin': '0.2',
            'color': '#4773EB',
            'fillcolor': '#5E85ED',
            'fontname': 'Arial',
            'fontcolor': 'black'
        },
        'user': {
            'shape': 'record',
            'style': 'filled',
            'margin': '0.2',
            'color': '#CAE188',
            'fillcolor': '#E3FE97',
            'fontname': 'Arial',
            'fontcolor': 'black'
        },
        
    }
    EDGE_STYLYING = {
        "fontname": "Arial"
    }
    
    def __init__(self, output_directory: str =None):
        
        self.dot = graphviz.Digraph()
        self.output_directory = output_directory
        
   
    def generate_call_flow_graph(self, nodes: list, number: str):
        self.dot.attr(name=f"calls_to_{number}", label=number, fontname='Arial',
                      fontcolor='white', rankdir="LR")
        
        # build nodes
        self.dot.node("Start", "Start", GraphvizModule.NODE_STYLING["start"])
        for n in nodes:
            
            try: 
                node_config = f"<extension> Extension: {n.extension} "
            except AttributeError:
                node_config = ""
            
            if isinstance(n, bre.User):
                node_config += f'| <name> Name: {n.first_name} {n.last_name}'
                self.dot.node(n.id, node_config, GraphvizModule.NODE_STYLING["user"])
                
            elif isinstance(n, bre.CallCenter) or isinstance(n, bre.HuntGroup):
                node_config += f"| <name> Name: {n.name} | <policy> Policy: {n.policy}"
                for i, a in enumerate(n.agents):
                    node_config += f"| <{a.id}> Agent {i+1}: {a.extension}"
                self.dot.node(n.service_user_id, node_config, 
                              GraphvizModule.NODE_STYLING["call_centre"] if isinstance(n, bre.CallCenter) 
                              else GraphvizModule.NODE_STYLING["hunt_group"])
            
            elif isinstance(n, bre.AutoAttendant):
                node_config += f"| <name> Name: {n.name}"
                for k in n.business_hours_menu.keys:
                    node_config += f"| <key{k.number}> Option {k.number}: {k.action}"
                    k.id = f"{n.service_user_id}:<key{k.number}>"
                self.dot.node(n.service_user_id, node_config, GraphvizModule.NODE_STYLING["auto_attendant"])
            
        # build edges
        for n in nodes:
            try:
                if n._start_node:
                    try:
                        self.dot.edge("Start", n.id)
                    except Exception:
                        self.dot.edge("Start", n.service_user_id)
            except AttributeError:
                # node is note the start node
                pass

            if isinstance(n, bre.User):
                if n.call_forwarding_always:
                    self._format_edge(n, n.call_forwarding_always, "CFA")
                if n.call_forwarding_busy:
                    self._format_edge(n, n.call_forwarding_busy, "CFB")    
                if n.call_forwarding_not_reachable:
                    self._format_edge(n, n.call_forwarding_not_reachable, "CFNR")
                        
            elif isinstance(n, bre.CallCenter):
                if n.bounced_calls_enabled:
                    self._format_edge(n, n.bounced_calls_transfer_to_phone_number, "BCT")
                if n.overflow_calls_action:
                    self._format_edge(n, n.overflow_calls_transfer_to_phone_number, "OF")
                if n.stranded_calls_action:
                    self._format_edge(n, n.stranded_calls_transfer_to_phone_number, "SCF")
                if n.stranded_call_unavailable_action:
                    self._format_edge(n, n.stranded_call_unavailable_transfer_to_phone_number, "USCF")
                        
            elif isinstance(n, bre.HuntGroup):
                if n.forward_after_timeout_enabled:
                    self._format_edge(n, n.no_answer_forward_to_phone_number, "NACF")
                if n.call_forward_not_reachable_enabled:
                    self._format_edge(n, n.call_forward_not_reachable_transfer_to_phone_number, "CFNR")
                        
            elif isinstance(n, bre.AutoAttendant):
                for key in n.business_hours_menu.keys:
                    if "Transfer" in key.action:
                        self._format_edge(key, key.phone_number, key.number)
                        
    
    def _format_edge(self, node_a: str, node_b: str, label: str):     
        if node_b is None:
            return None
        
        node_a_id = node_a.id if hasattr(node_a, 'id') else node_a.service_user_id
        node_b_id = node_b.id if hasattr(node_b, 'id') else node_b.service_user_id

        self.dot.edge(node_a_id, node_b_id, label, GraphvizModule.EDGE_STYLYING)

            
    def _save_graph(self, filename: str):
        self.dot.render(directory=self.output_directory, filename=filename,
                        format="svg", cleanup=True).replace('\\', '/')
    
        